# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import multiprocessing
from decimal import Decimal
from functools import partial

import numpy as np
import pandas as pd
from pandas import DataFrame

from supplychainpy._helpers import _data_cleansing
from supplychainpy._helpers._decorators import keyword_sniffer, log_this
from supplychainpy._helpers._enum_formats import FileFormats
from supplychainpy._helpers._enum_formats import PeriodFormats
from supplychainpy._helpers._data_cleansing import check_extension
from supplychainpy.bi.recommendation_generator import run_sku_recommendation, run_profile_recommendation
from supplychainpy.inventory import analyse_uncertain_demand
from supplychainpy.inventory import economic_order_quantity
from supplychainpy.inventory.abc_xyz import AbcXyz
import warnings

from supplychainpy.inventory.analyse_uncertain_demand import UncertainDemand

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

UNKNOWN = 'UNKNOWN'


@keyword_sniffer
def analyse(currency: str, z_value: Decimal = 1.28, reorder_cost: Decimal = 10, interval_length: int = 12,
            interval_type: str = 'month', **kwargs):
    """ Performs several types of common inventory analysis on the raw demand data. Including safety stock, reorder
    levels.

    Args:
        currency (Decimal):             Currency the raw data is stored in.
        z_value (Decimal):              Service level requirement
        reorder_cost (Decimal):         Cost to place a reorder based on the cost of operations over the period
        interval_length (Decimal):      The number of periods the demand data is grouped into e.g. 12 (months)
        interval_type (Decimal):        months, weeks,days, years, quarters.

    Keyword Args:
        **df(pd.DataFrame):             Pandas DataFrame containing the raw data in the correct format.
        **file_path (str) :             Path to csv or txt file containing formatted data.

    Returns:
        dict/pd.DataFrame:  Analysis of inventory profile.

    Examples:

        >>> from supplychainpy.model_inventory import analyse
        >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
        >>> from decimal import Decimal
        >>> analysed_data = analyse(file_path=ABS_FILE_PATH['COMPLETE_CSV_SM'],
        ...                         z_value=Decimal(1.28),
        ...                         reorder_cost=Decimal(400),
        ...                         retail_price=Decimal(455),
        ...                         file_type='csv',
        ...                         currency='USD')
        >>> analysis = [demand.orders_summary() for demand in analysed_data]
        >>> # Example using pandas DataFrame.
        >>> import pandas as pd
        >>> raw_df = pd.read_csv(ABS_FILE_PATH['COMPLETE_CSV_SM'])
        >>> analyse_kv = dict(
        ...     df=raw_df,
        ...     start=1,
        ...     interval_length=12,
        ...     interval_type='months',
        ...     z_value=Decimal(1.28),
        ...     reorder_cost=Decimal(400),
        ...     retail_price=Decimal(455),
        ...     file_type='csv',
        ...     currency='USD'
        ... )
        >>> analysis_df = analyse(**analyse_kv)
    """

    analysed_orders = UncertainDemand
    analysed_orders_collection = []
    pool = multiprocessing.Pool(4)
    collect_kwargs =[]

    try:
        if kwargs is not None:
            if kwargs.get('file_path') == UNKNOWN:
                converted_data = list(np.array(kwargs['df']))
                for i in converted_data:
                    sku_id, orders, unit_cost, lead_time, retail_price, quantity_on_hand, backlog = \
                        _unpack_row(row=i, interval_length=interval_length)
                    total_orders = 0
                    if quantity_on_hand is None:
                        quantity_on_hand = 0.0
                    for order in orders['demand']:
                        total_orders += int(order)
                    analysed_orders = _analyse_orders(orders=orders, sku_id=sku_id, lead_time=lead_time,
                                                      unit_cost=unit_cost, reorder_cost=reorder_cost,
                                                      z_value=z_value, retail_price=retail_price,
                                                      quantity_on_hand=quantity_on_hand, currency=currency,
                                                      total_orders=total_orders, backlog=backlog)
                    analysed_orders_collection.append(analysed_orders)
                AbcXyz(analysed_orders_collection)
                analysis_df = _convert_to_pandas_df(analysed_orders_collection)
                return analysis_df
            elif kwargs['file_path'] is not UNKNOWN:
                analysed_orders_collection = []
                item_list = _clean_file(file_type=kwargs['file_type'], file_path=kwargs['file_path'],
                                        interval_length=interval_length)
                for sku in item_list:
                    orders = {}
                    sku_id, unit_cost, lead_time, retail_price, quantity_on_hand, backlog = sku.get("sku_id","UNKNOWN_SKU"), \
                                                                                   sku.get("unit_cost"), \
                                                                                   sku.get("lead_time"), \
                                                                                   sku.get("retail_price"), \
                                                                                   sku.get("quantity_on_hand"), \
                                                                                   sku.get("backlog")
                    orders['demand'] = sku.get("demand")
                    total_orders = 0
                    if quantity_on_hand is None:
                        quantity_on_hand = 0.0
                    for order in orders['demand']:
                        total_orders += int(order)
                    kwargs = {'orders':orders,
                              'sku_id':sku_id,
                              'lead_time':lead_time,
                              'unit_cost':unit_cost,
                              'reorder_cost':reorder_cost,
                              'z_value':z_value,
                              'retail_price':retail_price,
                              'quantity_on_hand':quantity_on_hand,
                              'currency':currency,
                              'total_orders':total_orders,
                              'backlog': float(backlog)}
                    collect_kwargs.append(kwargs)
                analysed_orders = [_analyse_orders(
                                                x['orders'],
                                                x['sku_id'],
                                                x['lead_time'],
                                                x['unit_cost'],
                                                x['reorder_cost'],
                                                x['z_value'],
                                                x['retail_price'],
                                                x['quantity_on_hand'],
                                                x['currency'],
                                                x['total_orders'],
                                                x['backlog']) for x in collect_kwargs]

                    #mapfunc = partial(_analyse_orders, **kwargs)

                    #analysed_orders = pool.map(mapfunc, 0)
                for x in analysed_orders:
                    analysed_orders_collection.append(x)
                AbcXyz(analysed_orders_collection)
                # analysis = [i.orders_summary() for i in analysed_orders_collection]
                return analysed_orders_collection
            elif kwargs['raw_data']:
                if kwargs['raw_data'] > 2:
                    d = analyse_uncertain_demand.UncertainDemand(orders=kwargs['raw_data'], sku=kwargs['sku_id'],
                                                                 lead_time=kwargs['lead_time'],
                                                                 unit_cost=kwargs['unit_cost'],
                                                                 reorder_cost=reorder_cost,
                                                                 z_value=z_value,
                                                                 retail_price=kwargs['raw_data'],
                                                                 quantity_on_hand=kwargs['quantity_on_hand'],
                                                                 currency=currency)
                else:
                    raise ValueError("Data set too small. Please use a minimum of 3 entries.")
                return d.orders_summary_simple()

    except KeyError as e:
        print(e)


def _clean_file(file_type: str, file_path: str, interval_length: int) -> dict:
    """ Cleans-up csv and txt files and validates the format.

    Args:
        file_type (str):            'csv' or 'txt' file type.
        file_path (str) :           Absolute path to file.
        interval_length (int):      The number of periods included in raw data.

    Returns:
        dict:   Sanitised source data.

    """
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                item_list = (_data_cleansing.clean_orders_data_row(raw_data, interval_length))
                return item_list
        elif file_type == FileFormats.csv.name:
            with open(file_path, 'r') as raw_data:
                item_list = _data_cleansing.clean_orders_data_row_csv(raw_data, length=interval_length)
                return item_list
    else:
        incorrect_file = "Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter."
        raise Exception(incorrect_file)


def _unpack_row(row: list, interval_length: int) -> tuple:
    """ Retrieves data-points based on fixed hardcoded positions in source file.

    Args:
        row (list):                 Row of data from source file.
        interval_length (int):      The number of periods included in raw data.

    Returns:
        tuple:      Separate data points (orders, sku_id, unit_cost, lead_time, retail_price, quantity_on_hand, backlog)

    """
    orders = {}
    sku_id = row[0]
    orders['demand'] = list(row[1:interval_length + 1])
    unit_cost = row[interval_length + 1]
    lead_time = row[interval_length + 2]
    retail_price = row[interval_length + 3]
    quantity_on_hand = row[interval_length + 4]
    backlog = row[interval_length + 5]
    return sku_id, orders, unit_cost, lead_time, retail_price, quantity_on_hand, backlog


def _convert_to_pandas_df(analysis: list) -> DataFrame:
    """ Converts list of dicts to Pandas Dataframe with the correct orientation.

    Args:
        analysis (list):    Collection of analysed SKU data.

    Returns:
        DataFrame:     Converted DataFrame.

    """
    d = [i.orders_summary() for i in analysis]
    analysis_dict = {
        'sku': [i.get('sku') for i in d],
        'unit_cost': [i.get('unit_cost') for i in d],
        'quantity_on_hand': [i.get('quantity_on_hand') for i in d],
        'excess_stock': [i.get('excess_stock') for i in d],
        'shortages': [i.get('shortages') for i in d],
        'demand_variability': [i.get('demand_variability') for i in d],
        'currency': [i.get('currency') for i in d],
        'safety_stock': [i.get('safety_stock') for i in d],
        'average_orders': [i.get('average_orders') for i in d],
        'economic_order_quantity': [i.get('economic_order_quantity') for i in d],
        'standard_deviation': [i.get('standard_deviation') for i in d],
        'ABC_XYZ_Classification': [i.get('ABC_XYZ_Classification') for i in d],
        'economic_order_variable_cost': [i.get('economic_order_variable_cost') for i in d],
        'reorder_quantity': [i.get('reorder_quantity') for i in d],
        'total_orders': [i.get('total_orders') for i in d],
        'reorder_level': [i.get('reorder_level') for i in d],
        'revenue': [i.get('revenue') for i in d]
    }
    analysis_df = pd.DataFrame(analysis_dict,
                               columns=['sku', 'unit_cost', 'quantity_on_hand', 'excess_stock', 'shortages',
                                        'demand_variability', 'currency', 'safety_stock', 'average_orders',
                                        'economic_order_quantity', 'standard_deviation',
                                        'ABC_XYZ_Classification',
                                        'economic_order_variable_cost', 'reorder_quantity', 'total_orders',
                                        'reorder_level', 'revenue'])
    return analysis_df


def _analyse_orders(orders: dict, sku_id: str, lead_time: Decimal, unit_cost: Decimal, reorder_cost: Decimal,
                    z_value: Decimal, retail_price: Decimal, quantity_on_hand: Decimal, currency: str,
                    total_orders: float, backlog: Decimal) -> UncertainDemand:
    """ Performs several types of common inventory analysis on the raw demand data. Including safety stock, reorder
    levels.

    Args:
        orders (dict):
        sku_id (str):
        lead_time (Decimal):
        unit_cost (Decimal):
        reorder_cost (Decimal):
        z_value (Decimal):
        retail_price (Decimal):
        quantity_on_hand (Decimal):
        currency (str):
        total_orders (float):

    Returns:
        UncertainDemand:    UncertainDemand objects.

    """
    analysed_orders = analyse_uncertain_demand.UncertainDemand(orders=orders,
                                                               sku=sku_id,
                                                               lead_time=lead_time,
                                                               unit_cost=unit_cost,
                                                               reorder_cost=Decimal(reorder_cost),
                                                               z_value=Decimal(z_value),
                                                               retail_price=retail_price,
                                                               quantity_on_hand=quantity_on_hand,
                                                               currency=currency,
                                                               backlog=backlog
                                                               )
    average_orders = analysed_orders.average_orders
    reorder_quantity = analysed_orders.fixed_order_quantity
    eoq = economic_order_quantity.EconomicOrderQuantity(total_orders=float(total_orders),
                                                        reorder_quantity=float(reorder_quantity),
                                                        holding_cost=float(0.25),
                                                        reorder_cost=float(reorder_cost),
                                                        average_orders=average_orders,
                                                        unit_cost=float(unit_cost))
    analysed_orders.economic_order_qty = eoq.economic_order_quantity
    analysed_orders.economic_order_variable_cost = eoq.minimum_variable_cost

    return analysed_orders


# TODO-feature specify length of orders, start position and period (day, week, month, ...)
# in the analysis function specify service-level expected for safety stock, a default is specified in the class
def analyse_orders(data_set: dict, sku_id: str, lead_time: Decimal, unit_cost: Decimal, reorder_cost: Decimal,
                   z_value: Decimal, retail_price: Decimal, quantity_on_hand: Decimal, currency: str = 'USD') -> dict:
    """Analyse orders data for one sku using a dictionary.

    Analyses orders data for a single sku using the values in the data_set dict.

    Args:
        data_set (dict):            The orders data for a specified period.
        sku_id (str):               The unique id of the sku.
        lead_time (Decimal):        The average lead-time for the sku over the period represented by the data,
                                    in the same unit.
        unit_cost (Decimal):        The unit cost of the sku to the organisation.
        reorder_cost (Decimal):     The cost to place a reorder. This is usually the cost of the operation divided by
                                    number of purchase orders placed in the previous period.
        z_value (Decimal):          The service level required to calculate the safety stock.
        retail_price (Decimal):     The retail or standard price of the sku.
        quantity_on_hand (Decimal): The quantity currently on hand as of analysis or retrieving data set.
        currency (str):             Currency of source raw data.

    Returns:
        dict:       The summary of the analysis.

                    `{'ABC_XYZ_Classification': 'AX', 'reorder_quantity': '258', 'revenue': '2090910.44',
                    'average_order': '539', 'reorder_level': '813', 'economic_order_quantity': '277', 'sku': 'RR381-33',
                    'demand_variability': '0.052', 'economic_order_variable_cost': '29557.61',
                    'standard_deviation': '28', 'safety_stock': '51'}`

    Raises:
        ValueError: Dataset too small. Please use a minimum of 3 entries.


    Examples:
        >>> from supplychainpy.model_inventory import analyse_orders
        >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
        >>> from decimal import Decimal
        >>> yearly_demand = {'jan': 75, 'feb': 75, 'mar': 75, 'apr': 75, 'may': 75, 'jun': 75, 'jul': 25,
        ...                 'aug': 25, 'sep': 25, 'oct': 25, 'nov': 25, 'dec': 25}
        >>>
        >>> summary = analyse_orders(yearly_demand,
        ...                          sku_id='RX983-90',
        ...                          lead_time=Decimal(3),
        ...                          unit_cost=Decimal(50.99),
        ...                          reorder_cost=Decimal(400),
        ...                          z_value=Decimal(1.28),
        ...                          retail_price=Decimal(600),
        ...                          quantity_on_hand=Decimal(390))
    """
    if len(data_set) > 2:
        warnings.warn('The \'analyse_orders\' function has been deprecated. Please use the \'analyse\' function',
                      DeprecationWarning)
        d = analyse_uncertain_demand.UncertainDemand(orders=data_set, sku=sku_id, lead_time=lead_time,
                                                     unit_cost=unit_cost, reorder_cost=reorder_cost,
                                                     z_value=z_value, retail_price=retail_price,
                                                     quantity_on_hand=quantity_on_hand, currency=currency)
    else:
        raise ValueError('Data set too small. Please use a minimum of 3 entries.')
    return d.orders_summary_simple()


def analyse_orders_from_file_col(file_path, sku_id: str, lead_time: Decimal, unit_cost: Decimal,
                                 reorder_cost: Decimal, z_value: Decimal, retail_price: Decimal,
                                 file_type: str = FileFormats.text.name,
                                 period: str = PeriodFormats.months.name, currency='USD') -> dict:
    """ Analyse orders from file arranged in a single column.

    Analyses orders data for a single sku, using the values
    from a file arranged in columns.The data should be arranged in two columns, 1 for the period and the other for the
    corresponding data-point.

    Args:
        file_path (str):              The path to the file containing two columns of data, 1 period and 1 data-point for 1 sku.
        sku_id (str):                 The unique id of the sku.
        lead_time (Decimal):          The average lead-time for the sku over the period represented by the data, in the same unit.
        unit_cost (Decimal):          The unit cost of the sku to the organisation.
        reorder_cost (Decimal):       The cost to place a reorder. This is usually the cost of the operation divided by number of purchase orders placed in the previous period.
        z_value (Decimal):            The service level required to calculate the safety stock
        retail_price (Decimal):       The price at which the sku is retailed.
        file_type (str):              Type of 'file csv' or 'text'
        period (int):                 The period of time the data points are bucketed into.
        currency (str):               The currency of the source data.

    Returns:
        dict:   The summary of the analysis.

                `{'ABC_XYZ_Classification': 'AX', 'reorder_quantity': '258', 'revenue': '2090910.44',
                'average_order': '539', 'reorder_level': '813', 'economic_order_quantity': '277', 'sku': 'RR381-33',
                'demand_variability': '0.052', 'economic_order_variable_cost': '29557.61',
                'standard_deviation': '28', 'safety_stock': '51'}`

    Raises:
        Exception:  Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter.
        Exception:  Unspecified file type, Please specify 'csv' or 'text' for file_type parameter.

    Examples:

    >>> from decimal import Decimal
    >>> from supplychainpy.model_inventory import analyse_orders_from_file_col
    >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
    >>> # text file
    >>> RX9304_43_analysis_txt = analyse_orders_from_file_col(file_path=ABS_FILE_PATH['PARTIAL_COL_TXT_SM'],
    ...                                                   sku_id='RX9304-43',
    ...                                                   lead_time=Decimal(2),
    ...                                                   unit_cost=Decimal(400),
    ...                                                   reorder_cost=Decimal(45),
    ...                                                   z_value=Decimal(1.28),
    ...                                                   file_type='text',
    ...                                                   retail_price=Decimal(30))
    >>> #csv file
    >>> RX9304_43_analysis_csv = analyse_orders_from_file_col(ABS_FILE_PATH['PARTIAL_COL_CSV_SM'], 'RX9304-43',
    ...                                                     reorder_cost=Decimal(45),
    ...                                                     unit_cost=Decimal(400),
    ...                                                     lead_time=Decimal(45),
    ...                                                     z_value=Decimal(1.28),
    ...                                                     file_type="csv",
    ...                                                     retail_price=Decimal(30))


    """
    orders = {}
    file_type_processed = ''
    # warnings.warn('Analyse orders function has been deprecated. Please use the analyse function', DeprecationWarning)
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                orders = _data_cleansing.clean_orders_data_col_txt(raw_data)
            file_type_processed = 'txt'
        elif file_type == FileFormats.csv.name:
            with open(file_path) as raw_data:
                orders = _data_cleansing.clean_orders_data_col_csv(raw_data)
            file_type_processed = 'csv'
    else:
        raise Exception('Unspecified file type, Please specify \'csv\' or \'text\' for file_type parameter.')

    d = analyse_uncertain_demand.UncertainDemand(orders=orders, sku=sku_id, lead_time=lead_time,
                                                 unit_cost=unit_cost, reorder_cost=reorder_cost, z_value=z_value,
                                                 period=period, retail_price=retail_price, currency=currency)
    log.debug('Raw data columnar from {} analysed'.format(file_type_processed))

    return d.orders_summary()


def analyse_orders_from_file_row(file_path: str, z_value: Decimal, reorder_cost: Decimal, retail_price: Decimal,
                                 file_type: str = FileFormats.text.name,
                                 period: str = "month", length: int = 12, currency: str = 'USD') -> list:
    """ Analyse multiple SKUs from a file with data arranged by row.

    Args:
        file_path (file):       The path to the file containing two columns of data, 1 period and 1 data-point for 1 sku.
        reorder_cost (Decimal): The cost to place a reorder. This is usually the cost of the operation divided by number
                                of purchase orders placed in the previous period.
        z_value (Decimal):      The service level required to calculate the safety stock
        file_type (str):        Type of 'file csv' or 'text'
        period (str):           The period of time the data points are bucketed into.
        length (int):           The length of the period.

    Returns:
        list:       A list of summaries containint

    Raises:
        Exception:  Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter.
        Exception:  Unspecified file type, Please specify 'csv' or 'text' for file_type parameter.

    Examples:
    >>> from supplychainpy.model_inventory import analyse_orders_from_file_row
    >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
    >>> analysed_data = analyse_orders_from_file_row(file_path=ABS_FILE_PATH['COMPLETE_CSV_SM'],
    ...                                                     reorder_cost=Decimal(45),
    ...                                                     z_value=Decimal(1.28),
    ...                                                     file_type="csv",
    ...                                                     retail_price=Decimal(30),
    ...                                                     currency='USD')




    """
    warnings.warn('Analyse orders function has been deprecated. Please use the analyse function', DeprecationWarning)
    orders = {}
    analysed_orders_summary = []
    analysed_orders_collection = []
    item_list = {}
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                item_list = (_data_cleansing.clean_orders_data_row(raw_data, length=length))
        elif file_type == FileFormats.csv.name:
            with open(file_path, 'r') as raw_data:
                item_list = _data_cleansing.clean_orders_data_row_csv(raw_data, length=length)

    else:
        unspecified_file = "Unspecified file type, Please specify 'csv' or 'text' for file_type parameter."
        raise Exception(unspecified_file)
    # maybe use an iterable instead of unpacking for constructor
    try:
        for sku in item_list:
            sku_id = sku.get("sku_id")
            unit_cost = sku.get("unit_cost")
            lead_time = sku.get("lead_time")

            quantity_on_hand = sku.get("quantity_on_hand")

            if quantity_on_hand is None:
                quantity_on_hand = 0.0
            orders['demand'] = sku.get("demand")

            analysed_orders = analyse_uncertain_demand.UncertainDemand(orders=orders, sku=sku_id,
                                                                       lead_time=lead_time,
                                                                       unit_cost=unit_cost,
                                                                       reorder_cost=reorder_cost, z_value=z_value,
                                                                       retail_price=retail_price,
                                                                       quantity_on_hand=quantity_on_hand,
                                                                       currency=currency)
            analysed_orders_collection.append(analysed_orders)
            analysed_orders_summary.append(analysed_orders.orders_summary())
            orders = {}
            del analysed_orders
    except KeyError:
        print("The csv file is incorrectly formatted")
    return analysed_orders_summary


# TODO Remove hard coded holding cost and make it a parameter
def analyse_orders_abcxyz_from_file(file_path: str, z_value: Decimal, reorder_cost: Decimal,
                                    file_type: str = FileFormats.text.name,
                                    period: str = "month", length: int = 12, currency: str = 'USD') -> list:
    """
    Analyse orders data from file and returns ABCXYZ analysis

    Analyses orders data for a single sku, using the values from a file arranged in columns.The data should be arranged
    in two columns, 1 for the period and the other for the corresponding data-point.

    Args:
        file_path (str):        The path to the file containing two columns of data, 1 period and 1 data-point for 1 sku.
        reorder_cost (Decimal): The average lead-time for the sku over the period represented by the data,
                                in the same unit.
        length (int):           The number of periods in the data-ser referenced from the second column of the row
                                onwards.
        reorder_cost (Decimal): The cost to place a reorder. This is usually the cost of the operation divided by number
                                of purchase orders placed in the previous period.
        z_value (Decimal):      The service level required to calculate the safety stock
        file_type (str):        Type of 'file csv' or 'text'
        period (str):           The period of time the data points are bucketed into.
        currency (str):

    Returns:
        list:     An AbcXyz class object is returned.

    Raises:
        Exception:  Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter.
        Exception:  Unspecified file type, Please specify 'csv' or 'text' for file_type parameter.

    Examples:

    >>> from decimal import Decimal
    >>> from supplychainpy.model_inventory import analyse_orders_from_file_col
    >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
    >>> abc = analyse_orders_abcxyz_from_file(file_path=ABS_FILE_PATH['COMPLETE_CSV_SM'],
    ...                                                          z_value=Decimal(1.28),
    ...                                                          reorder_cost=Decimal(5000),
    ...                                                          file_type="csv")

    """
    warnings.warn('Analyse orders function has been deprecated. Please use the analyse function', DeprecationWarning)
    analysed_orders_collection = []
    item_list = {}
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                item_list = (_data_cleansing.clean_orders_data_row(raw_data, length))
        elif file_type == FileFormats.csv.name:
            with open(file_path) as raw_data:
                item_list = _data_cleansing.clean_orders_data_row_csv(raw_data, length=length)
    else:
        incorrect_file = "Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter."
        raise Exception(incorrect_file)

    for sku in item_list:
        orders = {}

        sku_id, unit_cost, lead_time, retail_price, quantity_on_hand = sku.get("sku_id", "UNKNOWN_SKU"), sku.get(
            "unit_cost"), sku.get(
            "lead_time"), sku.get("retail_price"), sku.get("quantity_on_hand")

        orders['demand'] = sku.get("demand")
        total_orders = 0
        if quantity_on_hand is None:
            quantity_on_hand = 0.0
        for order in orders['demand']:
            total_orders += int(order)

        analysed_orders = analyse_uncertain_demand.UncertainDemand(orders=orders,
                                                                   sku=sku_id,
                                                                   lead_time=lead_time,
                                                                   unit_cost=unit_cost,
                                                                   reorder_cost=Decimal(reorder_cost),
                                                                   z_value=Decimal(z_value),
                                                                   retail_price=retail_price,
                                                                   quantity_on_hand=quantity_on_hand, currency=currency)

        average_orders = analysed_orders.average_orders

        reorder_quantity = analysed_orders.fixed_order_quantity

        eoq = economic_order_quantity.EconomicOrderQuantity(total_orders=float(total_orders),
                                                            reorder_quantity=float(reorder_quantity),
                                                            holding_cost=float(0.25),
                                                            reorder_cost=float(reorder_cost),
                                                            average_orders=average_orders,
                                                            unit_cost=float(unit_cost))

        analysed_orders.economic_order_qty = eoq.economic_order_quantity
        analysed_orders.economic_order_variable_cost = eoq.minimum_variable_cost

        analysed_orders_collection.append(analysed_orders)

        AbcXyz(analysed_orders_collection)

    return analysed_orders_collection


def recommendations(analysed_orders: UncertainDemand, forecast: dict) -> dict:
    """ Generate Recommendations for each SKU and the inventory Profile.

    Args:
        analysed_orders (UncertainDemand):  UncertainDemand object of analysed orders.
        forecast (dict):                    Output from a Forecast.

    Returns:
        dict:   Returns recommendations for each sku and for the inventory profile.

    Examples:
    >>> from decimal import Decimal
    >>> from supplychainpy.sample_data.config import ABS_FILE_PATH
    >>> from supplychainpy.model_inventory import analyse
    >>> from supplychainpy.model_inventory import recommendations
    ...
    >>> analysed_order = analyse(file_path=ABS_FILE_PATH['COMPLETE_CSV_SM'],
    ...                                        z_value=Decimal(1.28),
    ...                                          reorder_cost=Decimal(5000),
    ...                                          file_type="csv", length=12,currency='USD')
    ...
    >>> holts_forecast = {analysis.sku_id: analysis.holts_trend_corrected_forecast for analysis in
    ...                     analyse(file_path=ABS_FILE_PATH['COMPLETE_CSV_SM'], z_value=Decimal(1.28),
    ...                                          reorder_cost=Decimal(5000), file_type="csv",
    ...                                          length=12,currency='USD')}
    ...
    >>> recommend = recommendations(analysed_orders=analysed_order, forecast=holts_forecast)
    """

    recommend = {'sku_recommendations': run_sku_recommendation(analysed_orders=analysed_orders, forecast=forecast),
                 'profile_recommendations': run_profile_recommendation(analysed_orders=analysed_orders, forecast=forecast),
                 }

    return recommend


def summarise(analysed_orders: UncertainDemand):
    pass

# the np method allows a numpy array to be used. This requires the specification of a period and length the data is
# supposed to cover. This method also allows the use of lead time arrays for calculating average leadtimes. There
# also be an analyse_orders_from_file_np. using the analyse_orders_np method to process each row.

# def analyse_orders_np(unit_cost: Decimal, period: np.array, z_value: Decimal, orders: np.array,
#                      lead_time_np: np.array = [],
#                      lead_time: Decimal = 0.00, length: int = 12) -> dict:
#    d = analyse_uncertain_demand.UncertainDemandNp(orders_np=orders, length=length, period=period)
#    d.print_period()



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class Machine:
    def __init__(self):
        # user specified features
        self.name = None
        self.maximum_throughput = None
        self.processing_time = None
        self.setup_time = None
        self.down_time = 0
        # generic features

    # remember to build all methods that control machine up-time behaviour
    def add_down_time(self):
        self.down_time += 1

    def __str__(self):
        return '{}'.format(self.name)

    def __repr__(self):
        return '{}'.format(self.name)


# TODO-feature singleton with global scope for production Job, holds sequence and cycle time

class MachineBuilder:
    """Abstract builder"""

    def __init__(self):
        self.machine = None

    def make_new_machine(self):
        self.machine = Machine()


class ProductionDirector:
    def __init__(self, machine_builder: MachineBuilder):
        self._machine_builder = machine_builder

    def construct_machine(self, name: str = "", maximum_throughput: int = 0, processing_time: int = 0):
        self._machine_builder.make_new_machine()
        self._machine_builder.machine.name = name
        self._machine_builder.machine.maximum_throughput = maximum_throughput
        self._machine_builder.machine.processing_time = processing_time

    def add_down_time(self):
        self._machine_builder.machine.down_time = self._machine_builder.machine.add_down_time()

    @property
    def machine(self) -> Machine():
        return self._machine_builder.machine


# use to add the most generic features to all machines built, allow the abstract class constructor to build the rest
# based on the client. Create different concrete builders for different types of machine with different generic features
class SimpleMachineBuilder(MachineBuilder):
    """Concrete builder, adds generic features to the machine"""
    pass



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import multiprocessing as mp
import concurrent.futures as cf
import pyximport

from concurrent.futures import ProcessPoolExecutor
from copy import deepcopy
from decimal import Decimal

from supplychainpy.simulations import monte_carlo
from supplychainpy._helpers._cpu_info import CoreCount

#pyximport.install()
from supplychainpy.simulations.sim_summary import summarize_monte_carlo, frame
from supplychainpy.simulations import monte_carlo_simulation

# def run_monte_carlo_mt(file_path: str, z_value: Decimal, runs: int, reorder_cost: Decimal,
#                       file_type: str, period_length: int = 12) -> list:
#    pool = ThreadPool(processes=4, )
#    async_results = pool.apply_async(run_monte_carlo,
#                                     (file_path, z_value, runs, reorder_cost, file_type, period_length,))
#
#    return_val = async_results.get()
#
#    return return_val
#
#
# def run(simulation_frame: list, period_length: int = 12):
#    pool = ThreadPool(processes=4, )
#    async_results = pool.apply_async(summarize_window, (simulation_frame, period_length,))
#
#    return_val = async_results.get()
#
#    return return_val



def run_transactions(random_demand: list, period_length: int, orders_analysis: list) -> list:
    transaction_report = []
    for sim_window in monte_carlo_simulation.simulation_window(random_normal_demand=random_demand,
                                                               period_length=period_length,
                                                               analysed_orders=orders_analysis):
        sim_dict = {"index": "{}".format(sim_window.index),
                    "period": "{}".format(sim_window.position),
                    "sku_id": sim_window.sku_id,
                    "opening_stock": "{}".format(round(sim_window.opening_stock)),
                    "demand": "{}".format(round(sim_window.demand)),
                    "closing_stock": "{}".format(round(sim_window.closing_stock)),
                    "delivery": "{}".format(round(sim_window.purchase_order_receipt_qty)),
                    "backlog": "{:.0f}".format(sim_window.backlog),
                    "po_raised": "{}".format(sim_window.po_number_raised),
                    "po_received": "{}".format(sim_window.po_number_received),
                    "po_quantity": "{:.0f}".format(int(sim_window.purchase_order_raised_qty)),
                    "shortage_cost": "{:.0f}".format(Decimal(sim_window.shortage_cost)),
                    "revenue": "{:.0f}".format(sim_window.revenue),
                    "quantity_sold": "{:0.0f}".format(sim_window.sold),
                    "shortage_units": "{:.0f}".format(sim_window.shortage_units),
                    "previous_backlog": "{:.0f}".format(sim_window.previous_backlog)}
        transaction_report.append(deepcopy([sim_dict]))
    return transaction_report

def random_demand(orders_analysis=None, period_length=None)->list:
    """"""
    simulation = monte_carlo.SetupMonteCarlo(analysed_orders=orders_analysis)
    r_demand = simulation.generate_normal_random_distribution(period_length=period_length)
    return r_demand


def get_demand(orders_analysis, period_length):
    demand = deepcopy(random_demand(orders_analysis= orders_analysis,  period_length= period_length))
    return demand


def run_monte_carlo(orders_analysis: list, runs: int, period_length: int = 12) -> list:
    """Runs monte carlo simulation.

    Generates random distribution for demand over period specified and creates a simulation window for opening_stock,
    demand, closing_stock, delivery and backlog for each sku in the data set. Creates a transaction summary window
    for inventory movements.

    Args:
        orders_analysis (list): list of UncertainDemand objects, containing the results of inventory analysis.
        period_length (int):    The number of periods define the simulation window e.g. 12 weeks, months etc.
        runs (int):             The number of times to run the simulation.

    Returns:
        list:       A list containing the transaction summary for each period.
                    [[{'closing_stock': '0', 'backlog': '240', 'po_quantity': '6009', 'po_raised': 'PO 49', 'revenue':
                    '1145799', 'opening_stock': '1700', 'index': '9', 'period': '1', 'shortage_units': '240.4612',
                    'po_received': '', 'delivery': '0', 'shortage_cost': '48621', 'demand': '1940',
                    'quantity_sold': '1699', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '240', 'backlog': '852', 'po_quantity': '6380', 'po_raised':
                    'PO 59', 'revenue': '0', 'opening_stock': '0', 'index': '9', 'period': '2',
                    'shortage_units': '611.5989', 'po_received': '', 'delivery': '0', 'shortage_cost': '123665',
                    'demand': '612', 'quantity_sold': '0', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '0', 'backlog': '383', 'po_quantity': '6152', 'po_raised':
                    'PO 69', 'revenue': '162070', 'opening_stock': '240', 'index': '9', 'period': '3',
                    'shortage_units': '383.4993', 'po_received': '', 'delivery': '0', 'shortage_cost': '77543',
                    'demand': '624', 'quantity_sold': '240', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '4463', 'backlog': '383', 'po_quantity': '1689', 'po_raised': 'PO 79',
                    'revenue': '2749508', 'opening_stock': '0', 'index': '9', 'period': '4',
                    'shortage_units': '0.000000', 'po_received': 'PO 49', 'delivery': '6010',
                    'shortage_cost': '0', 'demand': '1547', 'quantity_sold': '4079', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '8002', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '5393458', 'opening_stock': '4463', 'index': '9', 'period': '5',
                    'shortage_units': '0.000000', 'po_received': 'PO 59', 'delivery': '6381',
                    'shortage_cost': '0', 'demand': '2841', 'quantity_sold': '8002', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '13594', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '9162457', 'opening_stock': '8002', 'index': '9', 'period': '6',
                    'shortage_units': '0.000000', 'po_received': 'PO 69', 'delivery': '6153',
                    'shortage_cost': '0', 'demand': '561', 'quantity_sold': '13594', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '13144', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '8859056', 'opening_stock': '13594', 'index': '9', 'period': '7',
                    'shortage_units': '0.000000', 'po_received': 'PO 79', 'delivery': '1690',
                    'shortage_cost': '0', 'demand': '2140', 'quantity_sold': '13144', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '11666', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '7862688', 'opening_stock': '13144', 'index': '9', 'period': '8',
                    'shortage_units': '0.000000', 'po_received': '', 'delivery': '0', 'shortage_cost': '0',
                    'demand': '1478', 'quantity_sold': '11665', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '8542', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '5757540', 'opening_stock': '11666', 'index': '9', 'period': '9',
                    'shortage_units': '0.000000', 'po_received': '', 'delivery': '0', 'shortage_cost': '0',
                    'demand': '3123', 'quantity_sold': '8542', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '7832', 'backlog': '0', 'po_quantity': '0', 'po_raised': '',
                    'revenue': '5278665', 'opening_stock': '8542', 'index': '9', 'period': '10',
                    'shortage_units': '0.000000', 'po_received': '', 'delivery': '0', 'shortage_cost': '0',
                    'demand': '710', 'quantity_sold': '7831', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '5379', 'backlog': '0', 'po_quantity': '389', 'po_raised': 'PO 149',
                    'revenue': '3625634', 'opening_stock': '7832', 'index': '9', 'period': '11', 'shortage_units':
                    '0.000000', 'po_received': '', 'delivery': '0', 'shortage_cost': '0', 'demand': '2453',
                    'quantity_sold': '5379', 'sku_id': 'KR202-217'}]
                    [{'closing_stock': '3799', 'backlog': '0', 'po_quantity': '1969', 'po_raised': 'PO 159',
                    'revenue': '2560610', 'opening_stock': '5379', 'index': '9', 'period': '12',
                    'shortage_units': '0.000000', 'po_received': '', 'delivery': '0', 'shortage_cost': '0',
                    'demand': '1580', 'quantity_sold': '3799', 'sku_id': 'KR202-217'}],...]

    """

    transaction_report = []
    # add shortage cost,
    cpu_cores = CoreCount()
    cores = cpu_cores.core_count

    with ProcessPoolExecutor(max_workers=cores) as executor:
        transactions = {k: executor.submit(run_transactions, *(get_demand(orders_analysis=orders_analysis, period_length=period_length), period_length, orders_analysis)) for k in range(0, runs)}
        transactions_gen = {item : cf.as_completed(transactions[item]) for item in transactions}
        transaction_report = [transactions[item].result() for item in transactions_gen]
            #transaction_report.append(deepcopy(pool.apply(run_transactions, args=[random_demand, period_length, orders_analysis])))
    return transaction_report


def summarize_window(simulation_frame: list, period_length: int = 12):
    """ Summarizes the simulation window and provides the stockout percentage for each sku.

    Provides a summary   of inventory transactions, for the full period length for each run.

    Args:
       simulation_frame (list):     A collection of simulation windows.
       period_length (int):         The number of periods define the simulation window e.g. 12 weeks, months etc.

    Returns:
       list:    A list containing the results for each sku.[{'maximum_opening_stock': 13181.0,
                'standard_deviation_closing_stock': 4407.468727436796, 'standard_deviation_backlog': 485.2568835735928,
                'variance_backlog': 235474.24305555542, 'minimum_backlog': 0.0, 'maximum_closing_stock': 13181.0,
                'standard_deviation_shortage_cost': 348.8327422580442, 'variance_opening_stock': 20387450.520833332,
                'average_opening_stock': 6350.75, 'minimum_quantity_sold': 0.0,
                'standard_deviation_opening_stock': 4515.246451837744, 'stockout_percentage': 0.0833333358168602,
                'maximum_backlog': 1337.0, 'sku_id': 'KR202-240', 'standard_deviation_revenue': 4489.574661244527,
                'maximum_quantity_sold': 13181.0, 'total_quantity_sold': 76742.0,
                'variance_quantity_sold': 20156280.638888914, 'minimum_closing_stock': 0.0,
                'variance_closing_stock': 19425780.583333332, 'variance_shortage_units': 121684.28207126712,
                'index': '32', 'average_closing_stock': 6461.5, 'total_shortage_units': 1337.80962,
                'maximum_shortage_units': 1266.604, 'minimum_shortage_units': 0.0, 'average_backlog': 216.9166717529297,
                'minimum_opening_stock': 0.0}...]
    """

    summary = summarize_monte_carlo(simulation_frame, period_length)


    return summary


def summarise_frame(window_summary):
    """ Summarizes the simulation frame
    Generates a summary for run window, into 1 transaction summary for all the runs.

    Args:
        window_summary  (list): window summary for each period multiplied by the number of runs.

    Returns:
       list:    [{'variance_opening_stock': '5007', 'maximum_opening_stock': 13805, 'maximum_backlog': 2278.0,
                'minimum_quantity_sold': 0.0, 'average_backlog': '218', 'minimum_opening_stock': 0,
                'average_quantity_sold': '7204', 'maximum_quantity_sold': 13805.0, 'minimum_closing_stock': 0,
                'standard_deviation_backlog': '628', 'standard_deviation_quantity_sold': '4636',
                'maximum_closing_stock': 13805, 'standard_deviation_closing_stock': '4755',
                'average_shortage_units': '2278', 'service_level': '91.67','minimum_backlog': 0.0,
                'average_closing_stock': '7115', 'sku_id': 'KR202-240'},...]

    """

    frame_summary = frame(window_summary)
    return frame_summary


# TODO-feature optimise based on minimising excess or minimising shortages or service level
# TODO-optimisation move to cython or c++
def optimise_service_level(frame_summary: list, orders_analysis: list, service_level: float, runs: int,
                           percentage_increase: float) -> list:
    """ Optimises the safety stock for declared service level.

    Identifies which skus under performed (experiencing a service level lower than expected) after simulating
    transactions over a specific period. The safety stock for these items are increased and the analysis is monte carlo
    is run again.


    Args:
        frame_summary (list):           window summary for each period multiplied by the number of runs.
        orders_analysis (list):         prior analysis of orders data.
        service_level (list):           required service level as a percentage.
        runs (int):                     number of runs from previous
        percentage_increase (float):    the percentage increase required

    Returns:
       list:    Updated orders analysis with new saftey stock values based optimised from the simulation. The initial values
                from the analytical model.

    """
    # compare service levels, build list of skus below service level, change their safety stock increase by a percentage
    # run the monte carlo, keep doing until all skus' are above the requested service level
    # sim_optimise = optimise_sim(service_level=service_level, frame_summary=frame_summary, orders_analysis=orders_analysis)
    count_skus = True
    while count_skus:
        count_skus = False
        start = True
        if start:
            for sku in orders_analysis:
                for item in frame_summary:
                    if sku.sku_id == item['sku_id']:
                        if float(item['service_level']) <= service_level:
                            sku.safety_stock = round(Decimal(sku.safety_stock)) * Decimal(percentage_increase)
                            break
    count_skus = True
    while count_skus:
        count_skus = False
        sim = run_monte_carlo(orders_analysis=orders_analysis, runs=runs, period_length=12)

        sim_window = summarize_window(simulation_frame=sim, period_length=12)

        sim_frame = summarise_frame(sim_window)

        for sku in orders_analysis:
            for item in sim_frame:
                if sku.sku_id == item['sku_id']:
                    if float(item['service_level']) <= service_level:
                        count_skus = True
                        sku.safety_stock = round(Decimal(sku.safety_stock)) * Decimal(percentage_increase)
                        break

    return orders_analysis



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import uuid
from datetime import datetime

from sqlalchemy import (MetaData, create_engine, Table, Column, Integer, String, DateTime, ForeignKey)
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_ARCHIVE, ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG
from supplychainpy._helpers._db_connection import database_connection_uri
from supplychainpy._helpers._pickle_config import serialise_config

TRANSACTION_TYPES = ['archive', 'compress']

metadata = MetaData()

transaction_type = Table('transaction_type', metadata,
                         Column('type_id', Integer(), primary_key=True),
                         Column('type', String(255)),
                         Column('create_date', DateTime(), default=datetime.now)
                         )

logged_transaction = Table('logged_transaction', metadata,
                           Column('log_id', Integer(), primary_key=True),
                           Column('type_id', ForeignKey('transaction_type.type_id')),
                           Column('create_date', DateTime(), default=datetime.now)

                           )

compressed_files = Table('compressed_files', metadata,
                         Column('compressed_id', Integer(), primary_key=True),
                         Column('compressed_file_path', String(255), nullable=False),
                         Column('compressed_file_name', String(255), nullable=False),
                         Column('create_date', DateTime(), default=datetime.now)
                         )

csv_files = Table('csv_files', metadata,
                  Column('file_id', Integer(), primary_key=True),
                  Column('file_path', String(255), nullable=False),
                  Column('analysis_id', Integer(), nullable=False),
                  Column('file_name', String(255), nullable=False),
                  Column('transaction_id', ForeignKey('logged_transaction.log_id')),
                  Column('compressed_file_id', ForeignKey('compressed_files.compressed_id')),
                  Column('create_date', DateTime(), default=datetime.now),
                  Column('update_date', DateTime(), default=datetime.now, onupdate=datetime.now)
                  )


# print(ABS_FILE_PATH, uuid.uuid1())

def create_management_db():
    connection = create_engine(database_connection_uri(db='management', create=True))
    metadata.create_all(connection)
    for i in TRANSACTION_TYPES:
        ins = transaction_type.insert().values(
            type=i
        )

        connection.execute(ins)



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from kivy.animation import Animation
from kivy.app import App
from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.uix.label import Label
from kivy.properties import NumericProperty
from kivy.properties import ReferenceListProperty
from kivy.properties import ObjectProperty
from kivy.vector import Vector
from kivy.clock import Clock
from random import randint, random


class OrdersParcel(Widget):
    velocity_x = NumericProperty(0)
    velocity_y = NumericProperty(0)
    velocity = ReferenceListProperty(velocity_x, velocity_y)

    def move(self):
        self.pos = Vector(*self.velocity) + self.pos


class StockParcel(Widget):
    def receive_inventory(self, parcel):
        if self.collide_widget(parcel):
            vx, vy = parcel.velocity
            offset = (parcel.center_y - self.center_y) / (self.height / 2)
            bounced = Vector(-1 * vx, vy)
            vel = bounced * 1.1
            parcel.velocity = vel.x, vel.y + offset
            return True


class SimulationFramework(Widget):
    parcel = ObjectProperty(None)
    stock = ObjectProperty(None)

    def despatch_inventory(self, vel=(0, -4)):
        # use variables for size x, y. Increment for growth
        self.parcel.center = (self.width / 2, self.height / 2 - 58)
        print(self.parcel.center)
        self.stock.center = self.center
        self.parcel.velocity = vel
        # self.stock.center, self.stock.size = self.center, (80, 80)

    def set_globe(self, vel=(0, -4)):
        self.parcel.center = (self.width / 2, self.height)
        print(self.parcel.center)
        self.stock.center = self.center
        self.parcel.velocity = vel

    def update(self, dt):
        self.parcel.move()
        # if (self.parcel.y < self.y) or (self.parcel.top > self.top):
        #    self.parcel.velocity_y *= -1
        if self.parcel.y < self.y:
            self.set_globe(vel=(0, -4))
        if self.parcel.y > self.height:
            self.set_globe(vel=(0, -4))

        if self.stock.receive_inventory(self.parcel):
            self.update(self)
            self.despatch_inventory()


class SimulationGuiApp(App):
    def build(self):
        sim = SimulationFramework()
        sim.set_globe()
        Clock.schedule_interval(sim.update, 1.0 / 240.0)
        return sim

    globe = ObjectProperty(None)


if __name__ == '__main__':
    SimulationGuiApp().run()



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal
from operator import attrgetter


# TODO-feature Allow user to set the boundaries for classe classification
class AbcXyz:
    """ Applies abc xyz analysis on a collection of demand passed into the """

    __cumulative_total_revenue = Decimal(0)
    __percentage_revenue = Decimal(0)
    __a_class = "A"
    __b_class = "B"
    __c_class = "C"
    __x_class = "X"
    __y_class = "Y"
    __z_class = "Z"
    __abcxyz_summary = []

    def __init__(self, orders_collection: list):
        self.__orders = orders_collection
        self.__cumulative_total_revenue = self._cumulative_total()
        self.percentage_revenue()
        self.cumulative_percentage_revenue()
        self.abc_classification()
        self.xyz_classification()
        self.__abcxyz_summary = self.classification_summary()

    def __repr__(self):
        representation = "AX: {}, AY: {}, AZ: {} \nBX: {}, BY: {}, BZ: {} \nCX: {}, CY: {}, CZ: {}"
        return representation.format(len(self.__abcxyz_summary.get('AX')),
                                     len(self.__abcxyz_summary.get('AY')),
                                     len(self.__abcxyz_summary.get('AZ')),
                                     len(self.__abcxyz_summary.get('BX')),
                                     len(self.__abcxyz_summary.get('BY')),
                                     len(self.__abcxyz_summary.get('BZ')),
                                     len(self.__abcxyz_summary.get('CX')),
                                     len(self.__abcxyz_summary.get('CY')),
                                     len(self.__abcxyz_summary.get('CZ')))

    @property
    def ax(self):
        return self.__abcxyz_summary.get('AX')

    @property
    def ay(self):
        return self.__abcxyz_summary.get('AY')

    @property
    def az(self):
        return self.__abcxyz_summary.get('AZ')

    @property
    def bx(self):
        return self.__abcxyz_summary.get('BX')

    @property
    def by(self):
        return self.__abcxyz_summary.get('BY')

    @property
    def bz(self):
        return self.__abcxyz_summary.get('BZ')

    @property
    def cx(self):
        return self.__abcxyz_summary.get('CX')

    @property
    def cy(self):
        return self.__abcxyz_summary.get('CY')

    @property
    def cz(self):
        return self.__abcxyz_summary.get('CZ')


        # @property
        # def orders(self)->list:
        #     return self.__orders

    # @orders.setter
    # def orders(self, orders):
    #    self.__orders = orders

    @property
    def abcxyz_summary(self) -> list:
        return self.__abcxyz_summary

    @abcxyz_summary.setter
    def abcxyz_summary(self, abcxyz):
        self.__abcxyz_summary = abcxyz

    def _cumulative_total(self) -> Decimal:
        cumulative_total = Decimal(0)
        try:
            for sku in self.__orders:
                cumulative_total += Decimal(sku.revenue)
        except:
            for sku in self.__orders:
                cumulative_total += Decimal(sku.get('revenue'))
        return cumulative_total

    def percentage_revenue(self):
        try:
            for sku in self.__orders:
                sku.percentage_revenue = Decimal(sku.revenue) / Decimal(self.__cumulative_total_revenue)
        except:
            for sku in self.__orders:
                sku['percentage_revenue'] = Decimal(sku.get('revenue'))/ Decimal(self.__cumulative_total_revenue)

    def cumulative_percentage_revenue(self):
        previous_total = Decimal(0)
        for sku in sorted(self.__orders, key=attrgetter('revenue'), reverse=True):
            sku.cumulative_percentage = Decimal(sku.percentage_revenue) + previous_total
            previous_total = sku.percentage_revenue + previous_total

    def abc_classification(self):
        for sku in sorted(self.__orders, key=attrgetter('revenue'), reverse=True):
            if sku.cumulative_percentage <= .80:
                sku.abc_classification = self.__a_class
            elif .80 < sku.cumulative_percentage <= .90:
                sku.abc_classification = self.__b_class
            else:
                sku.abc_classification = self.__c_class

    def xyz_classification(self):
        for sku in self.__orders:
            if sku.demand_variability <= Decimal(0.20):
                sku.xyz_classification = self.__x_class
            elif Decimal(0.20) < sku.demand_variability <= Decimal(0.60):
                sku.xyz_classification = self.__y_class
            else:
                sku.xyz_classification = self.__z_class

    def classification_summary(self):

        ax_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'AX']

        ay_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'AY']

        az_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'AZ']

        bx_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'BX']

        by_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'BY']

        bz_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'BZ']

        cx_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'CX']

        cy_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'CY']

        cz_count = [sku.orders_summary() for sku in self.__orders if sku.abcxyz_classification == 'CZ']

        classification_matrix = {"AX": ax_count,
                                 "AY": ay_count,
                                 "AZ": az_count,
                                 "BX": bx_count,
                                 "BY": by_count,
                                 "BZ": bz_count,
                                 "CX": cx_count,
                                 "CY": cy_count,
                                 "CZ": cz_count}
        self.__orders = []
        return classification_matrix

        # ranking method returns the demand list of dictionaries back in order



from _decimal import Decimal

from behave import *
from pandas import DataFrame
import pandas as pd

from supplychainpy.model_inventory import analyse
from supplychainpy.sample_data.config import ABS_FILE_PATH

use_step_matcher("re")

@given("I extract data from a CSV")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context): 
    """
    raw_df = pd.read_csv(ABS_FILE_PATH['COMPLETE_CSV_SM'])
    context.response = raw_df


@when("I analyse raw data in a Pandas DataFrame")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context): 
    """
    analyse_kv = dict(
        df=context.response,
        start=1,
        interval_length=12,
        interval_type='months',
        z_value=Decimal(1.28),
        reorder_cost=Decimal(400),
        retail_price=Decimal(455),
        file_type='csv',
        currency='USD'
    )
    analysis_df = analyse(**analyse_kv)
    context.response = analysis_df



@then("It should return a DataFrame with Descriptive Statistics")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context): 
    """
    assert isinstance(context.response, DataFrame)


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sqlite3
import uuid
from shutil import copyfile

from sqlalchemy import create_engine
from supplychainpy._csv_management._controller._update_db import post_transaction
from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG, ABS_FILE_PATH_ARCHIVE, \
    ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG
from supplychainpy._helpers._db_connection import database_connection_uri
from supplychainpy._helpers._pickle_config import deserialise_config, serialise_config


class _Orchestrate:
    """Orchestrates the management of csv files used to generate the analysis uploaded to the database."""

    __file_path = ''

    def __init__(self):
        pass

    # Check config file for presence of existing management db if not create new one
    @staticmethod
    def check_for_db():
        """ Checks _archive directory for the presence of a csv_management database. If an csv_management database
            does not exist, then a new instance is created.

        Returns:
            bool:   Returns False indicating a database does exist and there is not a need to create one, or True
                    indicating a database does not exist and so one should be created.

        """
        config = deserialise_config(ABS_FILE_PATH_CSV_MANAGEMENT_CONFIG)
        f = list([i.name for i in os.scandir(ABS_FILE_PATH_ARCHIVE)])
        #print(f,'\n', config.get('database_name', 'UNKNOWN'))
        if config.get('database_name', 'UNKNOWN') in f:
            return False
        else:
            return True

    @staticmethod
    def copy_file():
        """Copies file used to load analysis and stores in archive folder for posterity.

        Returns:

        """
        # Retrieve file_name and path to the source data csv
        config_file = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
        file_name = config_file.get('file', 'UNKNOWN')
        file_path = config_file.get('database_path', 'UNKNOWN')
        print(file_path, file_name)
        if 'UNKNOWN' not in (file_path, file_name):
            try:
                # Copy data source file to archive
                copyfile('{}{}'.format(file_path, file_name),
                         '{}/{}-{}.csv'.format(ABS_FILE_PATH_ARCHIVE, 'data_file', uuid.uuid1()))
                config_file['file_path'] = '{}/{}-{}.csv'.format(ABS_FILE_PATH_ARCHIVE, 'data_file', uuid.uuid1())
                serialise_config(config_file, ABS_FILE_PATH_APPLICATION_CONFIG)

            except FileExistsError as e:
                print('The source file is not in the same location as the analysis database. Please place the file in '
                      'the same location and restart the process.')

    @staticmethod
    def update_database(analysis_id) -> bool:
        """Updates csv management database with

        Args:
            analysis_id (int):  The id for the analysis the data file corresponds to as given by the transaction table,
                                when running an analysis

        Returns:
            bool:  Success or failure for the updating of the csv_management database.

        """
        connection = create_engine(database_connection_uri(db='management', retrieve=True))
        return post_transaction(analysis_id=analysis_id, connection=connection)

    def archive_file(self):
        """ Compresses the files in the archive once the _archive directory reaches a specified size.

        Returns:
            bool:   Success or failure of archiving process

        """
        f = [i for i in os.scandir(ABS_FILE_PATH_ARCHIVE)]
        print(f[0])


if __name__ == '__main__':
    d = _Orchestrate()
    d.copy_file()
    d.update_database(analysis_id=1)



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal, getcontext, ROUND_HALF_UP

import pyximport

from supplychainpy.inventory import analyse_uncertain_demand

#pyximport.install()
from supplychainpy.inventory.eoq import minimum_variable_cost
from supplychainpy.inventory.eoq import economic_order_quantity


class EconomicOrderQuantity:
    __economic_order_quantity = Decimal(0)
    analyse_uncertain_demand.UncertainDemand.__reorder_cost = Decimal(0)
    __holding_cost = Decimal(0)
    __min_variable_cost = Decimal(0)
    __reorder_quantity = Decimal(0)
    __unit_cost = 0.00

    @property
    def minimum_variable_cost(self) -> Decimal:
        return self.__min_variable_cost

    @property
    def economic_order_quantity(self) -> Decimal:
        return self.__economic_order_quantity

    def __init__(self, reorder_quantity: float, holding_cost: float, reorder_cost: float, average_orders: float,
                 unit_cost: float, total_orders: float):
        getcontext().prec = 8
        getcontext().rounding = ROUND_HALF_UP
        self.__reorder_quantity = Decimal(reorder_quantity)
        self.__holding_cost = holding_cost
        self.__reorder_cost = reorder_cost
        self.__unit_cost = unit_cost
        self.__min_variable_cost = minimum_variable_cost(total_orders, reorder_cost, unit_cost, holding_cost)
        self.__economic_order_quantity = economic_order_quantity(total_orders, reorder_cost, unit_cost, holding_cost,
                                                                 reorder_quantity)





# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Generator

import numpy as np
import logging
import pyximport

from supplychainpy._helpers._decorators import log_this

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

#pyximport.install()
from supplychainpy.demand._squared_error import sum_squared_errors_i1, sum_squared_errors_i2, sum_squared_errors_i0


class Forecast:
    """

    """
    __simple_exponential_smoothing_forecast = {}

    # make keyword args
    def __init__(self, orders: list = None, average_orders: float = None, **kwargs):
        self.__weighted_moving_average = None
        self.__orders = orders
        self.__average_orders = sum([int(demand) for demand in self.__orders]) / len(self.__orders)
        self.__moving_average = []
        self.__total_orders = sum([int(demand) for demand in self.__orders])

    @property
    def total_orders(self):
        return self.__total_orders

    @property
    def moving_average(self) -> list:
        return self.__moving_average

    @moving_average.setter
    def moving_average(self, forecast: list):
        self.__moving_average = forecast

    @property
    def weighted_moving_average(self) -> list:
        return self.__weighted_moving_average

    # specify a start position and the forecast will build from this position
    def moving_average_forecast(self, average_period: int = 3, forecast_length: int = 2,
                                base_forecast: bool = False, start_position: int = 0) -> list:
        """ Generates a forecast from moving averages.

        Generate a forecast from moving averages for as many periods as specified.

        Args:
            average_period (int):       Number of periods to average.
            forecast_length (int):      Number of periods to forecast for.
            base_forecast (bool):       Start a moving average forecast from anywhere
                                        in the list. For use when evaluating a forecast.
            start_position (int):       Where to start the forecast in the list when

        Returns:
            list:       Returns a list of orders including the original and the forecast appended to the end.
                        The appended forecast is as long as specified by the forecast_length. For example

                        orders = [1, 3, 5, 67, 4, 65, 242, 50, 48, 24, 34, 20]
                        d = forecast_demand.Forecast(orders)
                        d.calculate_moving_average_forecast(forecast_length=3)
                        print(d.moving_average_forecast)

                            output:

                                [1, 3, 5, 67, 4, 65, 242, 50, 48, 24, 34, 20, 26, 27, 24]

                        orders = [1, 3, 5, 67, 4, 65, 242, 50, 48, 24, 34, 20]
                        d = forecast_demand.Forecast(orders)
                        d.calculate_moving_average_forecast(forecast_length=3, base_forecast=True, start_position=3)
                        print(d.moving_average_forecast):

                            output:

                                [3, 67, 65, 45, 60, 80]

        Raises:
            ValueError: Incorrect number of orders supplied. Please make sure you have enough orders to
                        calculate an average. The average_period is {}, while the
                        number of orders supplied is {}. The number of orders supplied should be equal
                        or greater than the average_period. Either decrease the average_period or
                        increase the start_position in the list.n
        """

        if base_forecast:
            start_period = start_position + average_period
            end_period = len(self.__orders)
            total_orders = 0
            moving_average = []
            if len(self.__orders[0:start_position]) < average_period:
                raise ValueError("Incorrect number of orders supplied. Please make sure you have enough orders to "
                                 "calculate an average. The average_period is {}, while the \n"
                                 "number of orders supplied is {}. The number of orders supplied should  be equal "
                                 "or greater than the average_period.\n Either decrease the average_period or "
                                 "increase the start_position in the list.".format(average_period, start_position))
            else:
                for i in self.__orders[0:start_position]:
                    moving_average.append(self.__orders[i])

            count = 0
            average_orders = 0.0
            while count < forecast_length:
                for items in moving_average[0:end_period]:
                    total_orders += items
                count += 1
                average_orders = total_orders / float(average_period)
                moving_average.append(round(average_orders))
                average_orders = 0.0
                total_orders = 0.0
                if count < 1:
                    start_period += len(moving_average) - average_period
                else:
                    start_period += 1

                end_period = len(moving_average)
            self.__moving_average = moving_average

        else:
            start_period = len(self.__orders) - average_period
            end_period = len(self.__orders)
            total_orders = 0
            moving_average = self.__orders
            count = 0
            average_orders = 0.0
            while count < forecast_length:
                for items in moving_average[start_period:end_period]:
                    total_orders += items
                count += 1
                average_orders = total_orders / float(average_period)
                moving_average.append(round(average_orders))
                average_orders = 0.0
                total_orders = 0.0
                if count < 1:
                    start_period += len(moving_average) - average_period
                else:
                    start_period += 1

                end_period = len(moving_average)
                self.__moving_average = moving_average
            return moving_average

    def weighted_moving_average_forecast(self, weights: list, average_period: int = 3,
                                         forecast_length: int = 3, base_forecast: bool = False,
                                         start_position=0) -> list:
        """ Generates a forecast from moving averages using user supplied weights.

        Generate a forecast from moving averages for as many periods as specified and adjusts the forecast based
        on the supplied weights.

        Args:
            average_period (int):       Number of periods to average.
            forecast_length (int):      Number of periods to forecast for.
            weights (list):             A list of weights that sum up to one.
            base_forecast(bool):        Start a moving average forecast from anywhere
                                        in the list. For use when evaluating a forecast.
            start_position(int):        Start position

        Returns:
            list:       Returns a list of orders including the original and the forecast appended to the end.
                        The appended forecast is as long as specified by the forecast_length. For example

                        orders = [1, 3, 5, 67, 4, 65, 242, 50, 48, 24, 34, 20]
                        d = forecast_demand.Forecast(orders)
                        d.calculate_moving_average_forecast(forecast_length=3)
                        print(d.moving_average_forecast)

                            output:

                                [1, 3, 5, 67, 4, 65, 242, 50, 48, 24, 34, 20, 13, 11, 7]

        Raises:
            ValueError: The weights should equal 1 and be as long as the average period (default 3).'
                        The supplied weights total {} and is {} members long. Please check the supplied
                        weights.'.format(sum(weights), len(weights)))
        """

        if sum(weights) != 1 or len(weights) != average_period:
            raise ValueError(
                'The weights should equal 1 and be as long as the average period (default 3).'
                ' The supplied weights total {} and is {} members long. Please check the supplied weights.'.format(
                    sum(weights), len(weights)))
        else:
            start_period = len(self.__orders) - average_period

        if base_forecast:
            start_period = start_position + average_period
            end_period = len(self.__orders)
            total_orders = 0
            weighted_moving_average = []
            if start_position + 1 < average_period:
                raise ValueError("Incorrect number of orders supplied. Please make sure you have enough orders to "
                                 "calculate an average. The average_period is {}, while the \n"
                                 "number of orders supplied is {}. The number of orders supplied should  be equal "
                                 "or greater than the average_period.\n Either decrease the average_period or "
                                 "increase the start_position in the list.".format(average_period, start_position))
            else:
                for i in self.__orders[0:start_position]:
                    weighted_moving_average.append(self.__orders[i])

            count = 0
            weight_count = 0
            while count < forecast_length:
                for items in weighted_moving_average[0:end_period]:
                    total_orders += items

                count += 1
                average_orders = (total_orders / float(average_period)) * weights[weight_count]
                weighted_moving_average.append(round(average_orders))
                total_orders = 0.0
                if count < 1:
                    start_period += len(weighted_moving_average) - average_period
                else:
                    start_period += 1
            self.__weighted_moving_average = weighted_moving_average
        else:
            end_period = len(self.__orders)
            total_orders = 0
            weighted_moving_average = self.__orders
            count = 0
            average_orders = 0.0
            while count < forecast_length:
                weight_count = 0
                for items in weighted_moving_average[start_period:end_period]:
                    total_orders += items
                average_orders = (total_orders / float(average_period)) * weights[weight_count]
                count += 1
                weighted_moving_average.append(round(average_orders))
                average_orders = 0.0
                total_orders = 0.0
                if count < 1:
                    start_period += len(weighted_moving_average) - average_period
                else:
                    start_period += 1

                end_period = len(weighted_moving_average)

        self.__weighted_moving_average = weighted_moving_average
        return weighted_moving_average

    # also use to calculate the MAD for all forecasting methods given a spcific length of order

    # TODO-feature fix base_forecast for correct period to period MAD calculation
    def mean_absolute_deviation(self, forecasts: list, base_forecast: bool = False, start_period: int = 3) -> np.array:

        """ calculates the mean absolute deviation (MAD) for a given forecast.

        Calculates the mean absolute deviation for a forecast. The forecast and

        Args:
            forecasts (list):       A previously calculated forecast.
            base_forecast (bool):   Start a moving average forecast from anywhere
                                    in the list.
            start_period (int):     The start period of the forecast.

        Returns:
            np.array

        Raises:
            ValueError:
        """

        if base_forecast:
            end_period = len(forecasts) - start_period
            forecast_array = np.array(forecasts[:end_period])
            orders_array = np.array(self.__orders[start_period: len(forecasts)])
            std_array = orders_array - forecast_array
            std_array = sum(abs(std_array)) / len(std_array)
        else:
            forecast_array = np.array(forecasts)
            orders_array = np.array(self.__orders)
            std_array = orders_array - forecast_array
            std_array = sum(abs(std_array)) / len(std_array)

        return std_array

    @log_this(logging.DEBUG, "Called to compute simple exponential smoothing.")
    def simple_exponential_smoothing(self, *alpha)->Generator[dict, None, None]:
        """ Generates forecast using simple exponential smoothing (SES).

        Args:
            alpha(args):    A list of smoothing level constants (alpha values)

        Returns:
            dict:           Forecast containing
                            {'demand': 165, 'level_estimates': 164.49748124123246, 'alpha': 0.7487406206162335,
                            'one_step_forecast': 163.0, 'forecast_error': 2.0, 'squared_error': 4.0, 't': 1}
         Examples:
                            forecast_demand = Forecast(orders)
                            alpha_values = [0.2, 0.3, 0.4, 0.5, 0.6]
                            ses_forecast = [forecast for forecast in forecast_demand.simple_exponential_smoothing(*alpha_values)]

       """

        for arg in alpha:
            forecast = {}
            current_level_estimate = self.__average_orders
            forecast.update({'alpha': arg,
                             't': 0,
                             'demand': 0,
                             'level_estimates': current_level_estimate,
                             'one_step_forecast': 0,
                             'forecast_error': 0,
                             'squared_error': 0})
            previous_level_estimate = current_level_estimate
            for index, demand in enumerate(tuple(self.__orders), 1):
                current_level_estimate = self._level_estimate(previous_level_estimate, arg, demand)
                yield {'alpha': arg,
                       't': index,
                       'demand': demand,
                       'level_estimates': current_level_estimate,
                       'one_step_forecast': previous_level_estimate,
                       'forecast_error': self._forecast_error(demand, previous_level_estimate),
                       'squared_error': self._forecast_error(demand, previous_level_estimate) ** 2
                       }
                previous_level_estimate = current_level_estimate


    @log_this(logging.DEBUG, "Called HTEC exponential smoothing.")
    def holts_trend_corrected_exponential_smoothing(self, alpha: float, gamma: float, intercept: float, slope: float):
        forecast = {}
        # log.debug('holts ')
        current_level_estimate = intercept
        forecast.update({'alpha': alpha,
                         'gamma': gamma,
                         't': 0,
                         'demand': 0,
                         'level_estimates': current_level_estimate,
                         'trend': slope,
                         'one_step_forecast': 0,
                         'forecast_error': 0,
                         'squared_error': 0

                         })
        previous_trend = slope
        previous_level_estimate = current_level_estimate
        for index, demand in enumerate(tuple(self.__orders), 1):
            log.debug('demand: {}'.format(demand))
            one_step = previous_level_estimate + previous_trend
            log.debug('one_step: {}'.format(one_step))
            forecast_error = self._forecast_error(demand, one_step)
            log.debug('forecast_error: {}'.format(forecast_error))
            current_trend = self._holts_trend(previous_trend, gamma, alpha, forecast_error)
            log.debug('trend: {}'.format(current_trend))
            current_level_estimate = self._level_estimate_holts_trend_corrected(previous_level_estimate,
                                                                                alpha,
                                                                                previous_trend,
                                                                                forecast_error)
            log.debug('current_level: {}'.format(current_level_estimate))
            squared_error = forecast_error ** 2
            yield {'alpha': alpha,
                   'gamma': gamma,
                   't': index,
                   'demand': demand,
                   'trend': current_trend,
                   'level_estimates': current_level_estimate,
                   'one_step_forecast': one_step,
                   'forecast_error': forecast_error,
                   'squared_error': squared_error
                   }
            log.debug('squared_error: {}'.format(squared_error))
            previous_level_estimate = current_level_estimate
            previous_trend = current_trend

    @staticmethod
    def holts_trend_corrected_forecast(forecast: list, forecast_length: int):
        """Creates a forecast for as many periods.

        Args:
            forecast:
            forecast_length:

        Returns:

        """
        end_of_forecast = len(forecast) - 1
        # print(forecast[end_of_forecast])

        new_forecast = []

        for i in range(forecast_length):
            demand_forecast = forecast[end_of_forecast]['level_estimates'] + i * forecast[end_of_forecast]['trend']
            new_forecast.append(demand_forecast)
        return new_forecast

    @staticmethod
    def simple_exponential_smoothing_forecast(forecast: list, forecast_length: int):
        end_of_forecast = len(forecast) - 1
        new_forecast = []

        for i in range(forecast_length):
            demand_forecast = forecast[end_of_forecast]['level_estimates']
            new_forecast.append(demand_forecast)
        return new_forecast

    @staticmethod
    def _holts_trend(previous_trend: float, gamma: float, alpha: float, current_forecast_error: float):
        return previous_trend + alpha * gamma * current_forecast_error

    @staticmethod
    def _level_estimate(lvl: float, smoothing_parameter: float, demand: int):
        return lvl + smoothing_parameter * (demand - lvl)

    @staticmethod
    def _level_estimate_holts_trend_corrected(previous_level: float, smoothing_parameter: float, previous_trend: float,
                                              forecast_error: float):
        return previous_level + previous_trend + smoothing_parameter * forecast_error

    @staticmethod
    def _forecast_error(demand: int, one_step_forecast: float):
        return float(demand) - one_step_forecast

    @staticmethod
    def sum_squared_errors(squared_error: list, smoothing_parameter: float) -> dict:
        return sum_squared_errors_i0(squared_error, smoothing_parameter)

    @staticmethod
    def sum_squared_errors_indi(squared_error: list, smoothing_parameter: float) -> dict:
        return sum_squared_errors_i2(squared_error, smoothing_parameter)

    @staticmethod
    def sum_squared_errors_indi_htces(squared_error: list, alpha: float, gamma: float) -> dict:
        return sum_squared_errors_i1(squared_error, alpha, gamma)

    @staticmethod
    def standard_error(sse: dict, orders_count, smoothing_parameter, df: int = 1) -> float:
        return (sse[smoothing_parameter] / (orders_count - df)) ** 0.5

    def mean_forecast_error(self):

        """

        Args:
            forecasts (int):        Number of periods to average.
            base_forecast (bool):   A list of weights that sum up to one.


        Returns:
            np.array

        Raises:
            ValueError:
        """

        pass

    def mean_aboslute_percentage_error_opt(self, forecast: list) -> int:
        """ Calculates the mean absolute percentage error for the optimised forecasts

        Args:
            forecast (list): forecast generated

        Returns:

        """
        sum_ape = sum([abs((i['demand'] - i['level_estimates']) / i['demand']) for i in forecast])
        mape = (sum_ape / len(forecast)) * 100
        return mape

    def optimise(self):

        """

        Args:
            forecasts (int):        Number of periods to average.
            base_forecast (bool):   A list of weights that sum up to one.


        Returns:
            np.array

        Raises:
            ValueError:
        """
        pass

    def linear_regression(self):

        """

        Args:
            line (int):        Number of periods to average.
            base_forecast (bool):   A list of weights that sum up to one.


        Returns:
            np.array

        Raises:
            ValueError:
        """
        pass

    def autoregressive(self):
        """

        Args:
            forecasts (int):        Number of periods to average.
            base_forecast (bool):   A list of weights that sum up to one.


        Returns:
            np.array

        Raises:
            ValueError:
        """
        pass


if __name__ == '__main__':
    orders = [165, 171, 147, 143, 164, 160, 152, 150, 159, 169, 173, 203, 169, 166, 162, 147, 188, 161, 162, 169, 185,
              188, 200, 229, 189, 218, 185, 199, 210, 193, 211, 208, 216, 218, 264, 304]

    total_orders = 0
    avg_orders = 0
    for order in orders[:12]:
        total_orders += order

    avg_orders = total_orders / 12
    f = Forecast(orders, avg_orders)
    alpha = [0.2, 0.3, 0.4, 0.5, 0.6]
    s = [i for i in f.simple_exponential_smoothing(*alpha)]

    sum_squared_error = f.sum_squared_errors(s, 0.5)
    # print(sum_squared_error)

    standard_error = f.standard_error(sum_squared_error, len(orders), smoothing_parameter=0.5)
    # print(standard_error)



from supplychainpy.reporting.blueprints.dashboard.views import (dashboard,
                                                                sku_detail,
                                                                get_classification_summary,
                                                                abxyz,top_shortages,
                                                                top_excess,
                                                                sku,
                                                                total_inventory,
                                                                currency)



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from copy import deepcopy

# from pyximport import pyximport
# pyximport.install()
from supplychainpy._helpers._decorators import log_this
from supplychainpy.demand._forecast_demand import Forecast
from supplychainpy._helpers import _data_cleansing
from supplychainpy._helpers._data_cleansing import check_extension
from supplychainpy._helpers._enum_formats import FileFormats
from supplychainpy.demand._evo_algo import OptimiseSmoothingLevelGeneticAlgorithm
from supplychainpy.demand.regression import LinearRegression

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

UNKNOWN = "UNKNOWN"

@log_this(logging.INFO, "Called public api for simple_exponential_smoothing_forecast.")
def simple_exponential_smoothing_forecast(demand: list, smoothing_level_constant: float = 0.5, optimise: bool = True,
                                          forecast_length: int = 5, initial_estimate_period: int = 6,
                                           **kwargs) -> dict:
    """ Performs a simple exponential smoothing forecast on historical demand.

    Args:
        forecast_length (int):              Number of periods to extend the forecast.
        demand (list):                      Original historical demand.
        smoothing_level_constant (float):   Alpha value
        initial_estimate_period (int):      Number of period to use to derive an average for the initial estimate.
        **ds (pd.DataFrame):                Data frame with raw data.
        **optimise (bool)                   Optimisation flag for exponential smoothing forecast.

    Returns:
        dict: Simple exponential forecast

    Examples:

    >>> from supplychainpy.model_demand import simple_exponential_smoothing_forecast
    >>> orders = [165, 171, 147, 143, 164, 160, 152, 150, 159, 169, 173, 203, 169, 166, 162, 147, 188, 161, 162,
    ...           169, 185, 188, 200, 229, 189, 218, 185, 199, 210, 193, 211, 208, 216, 218, 264, 304]
    >>> ses = simple_exponential_smoothing_forecast(demand=orders, alpha=0.5, forecast_length=6, initial_period=18)


    """
    try:
        ds = kwargs.get('ds', 'UNKNOWN')
        if ds is not UNKNOWN:
            orders = list(kwargs.get('ds', "UNKNOWN"))
        else:
            orders = [int(i) for i in demand]
            forecast_demand = Forecast(orders)

            log.log(logging.INFO, "Started simple exponential smoothing")
            # optimise, population_size, genome_length, mutation_probability, recombination_types
            if optimise:
                log.log(logging.INFO, "Using solver version to find alpha.")
                ses_forecast = [i for i in
                                forecast_demand.simple_exponential_smoothing(*(smoothing_level_constant,))]
                sum_squared_error = forecast_demand.sum_squared_errors(ses_forecast, smoothing_level_constant)
                standard_error = forecast_demand.standard_error(sum_squared_error, len(orders),
                                                                smoothing_level_constant)
                total_orders = 0
                for order in orders[:initial_estimate_period]:
                    total_orders += order
                avg_orders = total_orders // initial_estimate_period
                evo_mod = OptimiseSmoothingLevelGeneticAlgorithm(orders=orders,
                                                                 average_order=avg_orders,
                                                                 smoothing_level=smoothing_level_constant,
                                                                 population_size=10,
                                                                 standard_error=standard_error,
                                                                 recombination_type='single_point')

                ses_evo_forecast = evo_mod.simple_exponential_smoothing_evo(
                    smoothing_level_constant=smoothing_level_constant,
                    initial_estimate_period=initial_estimate_period)
                return ses_evo_forecast
            else:
                orders = [int(i) for i in demand]
                return _ses_forecast(smoothing_level_constant=smoothing_level_constant,
                                     forecast_demand=forecast_demand,
                                     forecast_length=forecast_length, orders_length=len(orders))


    except TypeError as e:
        if demand is None:
            print("Please supply a list of demand values. Use the keyword \'demand=\'\n{}".format(e))
    except OSError as e:
        print(e)


def _ses_forecast(smoothing_level_constant: float, forecast_demand: Forecast, forecast_length: int, orders_length: int) -> dict:
    """ Private function for executing the simple exponential smoothing forecast.
    """
    forecast_breakdown = list([i for i in forecast_demand.simple_exponential_smoothing(smoothing_level_constant)])
    ape = LinearRegression(forecast_breakdown)
    mape = forecast_demand.mean_aboslute_percentage_error_opt(forecast_breakdown)
    stats = ape.least_squared_error()
    simple_forecast = forecast_demand.simple_exponential_smoothing_forecast(forecast=forecast_breakdown,
                                                                            forecast_length=forecast_length)

    sum_squared_error = forecast_demand.sum_squared_errors(simple_forecast, smoothing_level_constant)
    standard_error = forecast_demand.standard_error(sum_squared_error, orders_length,
                                                    smoothing_level_constant)
    regression_line = _regr_ln(stats=stats)
    log.log(logging.WARNING,
            "A STANDARD simple exponential smoothing forecast has been completed.")
    return {'forecast_breakdown': forecast_breakdown, 'mape': mape, 'statistics': stats,
            'forecast': simple_forecast, 'alpha': smoothing_level_constant,
            'standard_error': standard_error,
            'regression': [i for i in regression_line.get('regression')]}


def simple_exponential_smoothing_forecast_from_file(file_path: str, file_type: str, length: int,
                                                    smoothing_level_constant: float = 0.5, forecast_length=5,
                                                    optimise: bool= True) -> dict:
    """Performs a simple exponential smoothing forecast on historical demand from file using a generator.

    Args:
        file_path (str): Path to source file.
        file_type (str): File type.
        length (int): Number of periods in data source (Jan..Dec = 12).
        smoothing_level_constant (int): Alpha value for forecast.
        forecast_length (int):  Number of periods to extend the forecast.
        optimise (bool): Optimisation flag for exponential smoothing forecast.

    Returns:
        dict: Simple exponential forecast.
        
    
    """
    item_list = {}
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                item_list = (_data_cleansing.clean_orders_data_row(file=raw_data, length=length))
        elif file_type == FileFormats.csv.name:
            with open(file_path) as raw_data:
                item_list = _data_cleansing.clean_orders_data_row_csv(file=raw_data, length=length)
    else:
        incorrect_file = "Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter."
        raise Exception(incorrect_file)

    for sku in item_list:

        sku_id, unit_cost, lead_time, retail_price, quantity_on_hand = sku.get("sku_id"), sku.get("unit_cost"), sku.get(
            "lead_time"), sku.get("retail_price"), sku.get("quantity_on_hand")
        orders = [int(i) for i in sku.get("demand")]
        if optimise:
            log.log(logging.WARNING,
                    "An OPTIMISED simple exponential smoothing forecast has been completed for SKU {}.".format(sku_id))
            yield {sku_id: simple_exponential_smoothing_forecast(demand=orders,
                                                                 forecast_length=forecast_length,
                                                                 smoothing_level_constant=smoothing_level_constant,
                                                                 optimise=True)}
        else:
            log.log(logging.WARNING,
                    "A STANDARD simple exponential smoothing forecast has been completed for SKU {}.".format(sku_id))
            yield {sku_id: simple_exponential_smoothing_forecast(demand=orders,
                                                                 forecast_length=forecast_length,
                                                                 smoothing_level_constant=smoothing_level_constant)}


def holts_trend_corrected_exponential_smoothing_forecast(demand: list, alpha: float, gamma: float,
                                                         forecast_length: int = 4, initial_period: int = 6,
                                                         optimise: bool = True) -> dict:
    """ Performs a holt's trend corrected exponential smoothing forecast on known demand

    Args:
        demand (list):  Original historical demand.
        alpha (float): smoothing constant
        gamma (float):
        forecast_length (int):
        initial_period (int):
        optimise (bool):    Flag for using solver. Default is set to True. 

    Returns:
        dict: Simple exponential forecast.

    """
    if optimise:
        total_orders = 0
        for order in demand[:initial_period]:
            total_orders += order
        avg_orders = total_orders // initial_period
        forecast_demand = Forecast(demand)
        processed_demand = [{'t': index, 'demand': order} for index, order in enumerate(demand, 1)]
        stats = LinearRegression(processed_demand)
        log_stats = stats.least_squared_error(slice_end=6)
        htces_forecast = [i for i in
                          forecast_demand.holts_trend_corrected_exponential_smoothing(alpha=alpha, gamma=gamma,
                                                                                      intercept=log_stats.get(
                                                                                          'intercept'),
                                                                                      slope=log_stats.get(
                                                                                          'slope'))]
        sum_squared_error = forecast_demand.sum_squared_errors_indi_htces(squared_error=[htces_forecast],
                                                                          alpha=alpha, gamma=gamma)
        standard_error = forecast_demand.standard_error(sum_squared_error, len(demand), (alpha, gamma), 2)
        evo_mod = OptimiseSmoothingLevelGeneticAlgorithm(orders=demand,
                                                         average_order=avg_orders,
                                                         population_size=10,
                                                         standard_error=standard_error,
                                                         recombination_type='single_point')
        optimal_alpha = evo_mod.initial_population(individual_type='htces')
        log.log(logging.WARNING,
                'An optimal alpha {} and optimal gamma {} have been found.'.format(optimal_alpha[1][0],
                                                                                   optimal_alpha[1][1]))
        htces_forecast = [i for i in
                          forecast_demand.holts_trend_corrected_exponential_smoothing(alpha=optimal_alpha[1][0],
                                                                                      gamma=optimal_alpha[1][1],
                                                                                      intercept=log_stats.get(
                                                                                          'intercept'),
                                                                                      slope=log_stats.get('slope'))]
        holts_forecast = forecast_demand.holts_trend_corrected_forecast(forecast=htces_forecast,
                                                                        forecast_length=forecast_length)
        log.log(logging.INFO, 'An OPTIMAL Holts trend exponential smoothing forecast has been generated.')
        sum_squared_error_opt = forecast_demand.sum_squared_errors_indi_htces(squared_error=[htces_forecast],
                                                                              alpha=optimal_alpha[1][0],
                                                                              gamma=optimal_alpha[1][1])
        standard_error_opt = forecast_demand.standard_error(sum_squared_error_opt, len(demand),
                                                            (optimal_alpha[1][0], optimal_alpha[1][1]), 2)

        ape = LinearRegression(htces_forecast)
        mape = forecast_demand.mean_aboslute_percentage_error_opt(htces_forecast)
        stats = ape.least_squared_error()
        regression_line = deepcopy(_regr_ln(stats=stats))
        return {'forecast_breakdown': htces_forecast, 'forecast': holts_forecast, 'mape': mape, 'statistics': stats,
                'optimal_alpha': optimal_alpha[1][0],
                'optimal_gamma': optimal_alpha[1][1],
                'SSE': sum_squared_error_opt,
                'standard_error': standard_error_opt,
                'original_standard_error': standard_error,
                'regression': [i for i in regression_line.get('regression')]}

    else:

        forecast_demand = Forecast(demand)
        processed_demand = [{'t': index, 'demand': order} for index, order in enumerate(demand, 1)]
        stats = LinearRegression(processed_demand)
        log_stats = stats.least_squared_error(slice_end=6)
        htces_forecast = [i for i in
                          forecast_demand.holts_trend_corrected_exponential_smoothing(alpha=alpha, gamma=gamma,
                                                                                      intercept=log_stats.get(
                                                                                          'intercept'),
                                                                                      slope=log_stats.get(
                                                                                          'slope'))]

        holts_forecast = forecast_demand.holts_trend_corrected_forecast(forecast=htces_forecast,
                                                                        forecast_length=forecast_length)
        log.log(logging.INFO, 'A STANDARD Holts trend exponential smoothing forecast has been generated.')
        sum_squared_error = forecast_demand.sum_squared_errors_indi_htces(squared_error=[htces_forecast],
                                                                          alpha=alpha, gamma=gamma)
        ape = LinearRegression(htces_forecast)
        mape = forecast_demand.mean_aboslute_percentage_error_opt(htces_forecast)
        stats = ape.least_squared_error()
        regression_line = _regr_ln(stats=stats)
        log.log(logging.WARNING, "A STANDARD Holts trend exponential smoothing forecast has been completed.")
        return {'forecast_breakdown': htces_forecast,
                'forecast': holts_forecast,
                'mape': mape,
                'statistics': stats,
                'sum_squared_errors': sum_squared_error,
                'regression': [i for i in regression_line.get('regression')]}


def holts_trend_corrected_exponential_smoothing_forecast_from_file(file_path: str, file_type: str, length: int,
                                                                   alpha: float, gamma: float, **kwargs) -> dict:
    """
    
    Args:
        file_path: Path to source file.
        file_type: File type.
        length: Number of periods in data source (Jan..Dec = 12).
        alpha: 
        gamma: 
        **kwargs: 

    Returns:

    """
    item_list = {}
    if check_extension(file_path=file_path, file_type=file_type):
        if file_type == FileFormats.text.name:
            with open(file_path, 'r') as raw_data:
                item_list = (_data_cleansing.clean_orders_data_row(raw_data, length))
        elif file_type == FileFormats.csv.name:
            with open(file_path) as raw_data:
                item_list = _data_cleansing.clean_orders_data_row_csv(raw_data, length=length)
    else:
        incorrect_file = "Incorrect file type specified. Please specify 'csv' or 'text' for the file_type parameter."
        raise Exception(incorrect_file)

    for sku in item_list:
        sku_id, unit_cost, lead_time, retail_price, quantity_on_hand = sku.get("sku_id"), sku.get("unit_cost"), sku.get(
            "lead_time"), sku.get("retail_price"), sku.get("quantity_on_hand")

        orders = [int(i) for i in sku.get("demand")]

        if kwargs['optimise']:
            log.log(logging.WARNING,
                    "An OPTIMISED Holts trend exponential smoothing forecast has been completed for SKU {}.".format(
                        sku_id))
            yield {sku_id: holts_trend_corrected_exponential_smoothing_forecast(demand=orders, alpha=alpha, gamma=gamma,
                                                                                forecast_length=4, initial_period=18,
                                                                                optimise=True)}

        else:
            log.log(logging.WARNING,
                    "A STANDARD Holts trend exponential smoothing forecast has been complete for SKU {}.".format(
                        sku_id))
            yield {sku_id: holts_trend_corrected_exponential_smoothing_forecast(demand=orders, alpha=alpha, gamma=gamma,
                                                                                forecast_length=4, initial_period=18,
                                                                                optimise=False)}


def _regr_ln(stats: dict) -> dict:
    """Calculates regression"""
    regr = {'regression': [(stats.get('slope') * i) + stats.get('intercept') for i in range(0, 12)]}
    return regr



import flask
from flask import Blueprint
from flask import Flask

from supplychainpy.reporting.blueprints.models import InventoryAnalysis, MasterSkuList, \
    Recommendations, ProfileRecommendation
from supplychainpy.reporting.config.settings import ProdConfig
from supplychainpy.reporting.extensions import db

recommendations_blueprint = Blueprint('recommendations', __name__, template_folder='templates')


app = Flask(__name__)
app.config.from_object(ProdConfig)

@recommendations_blueprint.route('/recommended/<string:sku_id>', methods=['GET'])
def recommended(sku_id: str = None):
    """The route to the rest api for retrieving recommendations for a SKU.

    Args:
        sku_id:     The SKU ID for a product.
    Returns:

    """
    """Rest api for sku level recommendations."""
    sku = db.session.query(MasterSkuList).filter(MasterSkuList.id == sku_id).first()
    inventory = db.session.query(InventoryAnalysis.id).filter(InventoryAnalysis.sku_id == sku.id).first()
    recommend = db.session.query(Recommendations.statement).filter(Recommendations.analysis_id == inventory.id).all()
    return flask.jsonify(json_list=recommend)


@recommendations_blueprint.route('/feed', methods=['GET'])
def recommendations():
    recommend = db.session.query(Recommendations).all()
    profile = db.session.query(ProfileRecommendation).all()
    inventory = db.session.query(InventoryAnalysis).all()
    return flask.render_template('feed.html', inventory=inventory, profile=profile, recommendations=recommend)




# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import random
import uuid

from supplychainpy._helpers._codes import Currency
from supplychainpy._helpers._db_connection import database_connection_uri
from supplychainpy._helpers._decorators import strip_punctuation
from supplychainpy.bi._recommendations import ResponseSingleton
from supplychainpy.bot._controller import master_sku_list, excess_controller, shortage_controller, revenue_controller, \
    inventory_turns_controller, average_orders_controller, currency_symbol_controller, classification_controller, \
    safety_stock_controller, reorder_level_controller
from supplychainpy.bot._helpers import get_master_sku_list, _unpack_pos, _find_pronoun, filter_pos, filter_pos_type

from textblob import TextBlob
from textblob import Word


class DashStates:
    _MIN = ('least', 'lowest', 'smallest', 'fewest')
    _MAX = ('biggest', 'most', 'highest', 'largest')
    _END_STATE = 'response'
    _EMPTY = 'EMPTY'
    _TRANSITION_STATES = {
        'PRONOUN_STATE': 'pronoun',
        'GREETING_STATE': 'greeting',
        'DECONSTRUCTION_STATE': 'deconstruction',
        'RANDOM_STATE': 'random_utterance',
        'RESPONSE_STATE': 'response'
    }
    _NOUN_SINGULAR = [
        "{}!",
        "Again {}.It's probably time for a coffee!",
        "{} may be hard, don't let the details get you down.",
        "{} we will be running the show."
    ]

    KEYWORDS = ('average', 'retail', 'economic')

    ANALYSIS_KEYWORDS = ("safety stock", "reorder level","excess", "shortage", "revenue", "inventory turns", "average order", "classification")

    SALUTATION_KEYWORDS = ("hi", "hello", "wassup", "sup", "what's up", "ello", "how's tricks?")

    SALUTATION_RESPONSES = ["hi", "hello", "how's tricks?"]

    UNKNOWN_UTTERANCE = ["Not sure what you are saying!", "I don't understand!"]

    _SEMANTIC_DICTIONARY = [
        ('question', ('biggest', 'smallest'), (('WP', 0), ('VBZ', 1), ('DT', 2), ('JJS', 3), ('NN', 4))),
        ('question', ('biggest', 'smallest'), (('WDT', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5))),
        ('question', ('biggest', 'smallest'), (('WDT', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NNS', 5))),
        ('question', ('shortage', 'revenue', 'excess', 'classification'), (('WP', 0), ('VBZ', 1), ('DT', 2), ('NN', 3), ('IN', 4), ('NNP', 5))),
        ('question', ('equal', 'is'), (('WP', 0), ('VBZ', 1), ('DT', 2), ('NN', 3), ('IN', 4), ('NNP', 5))),
        ('question', ('biggest', 'smallest'), (('WDT', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5))),
        ('question', ('most', 'fewest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NNS', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NN', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NN', 6))),

        ('question', ('biggest', 'smallest'), (('WDT', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NN', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('JJ', 5), ('NNS', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('JJ', 5), ('NN', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('JJ', 5), ('NN', 6))),
        ('question', ('biggest', 'smallest'), (('WDT', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('JJ', 5), ('NN', 6))),
        ('question', ('biggest', 'smallest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('JJ', 5), ('NNS', 6))),
        ('question', ('most', 'fewest'), (('NNP', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NNS', 6))),
        ('question', ('most', 'fewest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('JJS', 4), ('NN', 5), ('NNS', 6))),
        ('question', ('most', 'fewest'), (('NNP', 0), ('NNP', 1), ('VBZ', 2), ('DT', 3), ('RBS', 4), ('JJ', 5), ('NNS', 6))),
        ('question', ('most', 'fewest'), (('NNP', 0), ('NN', 1), ('VBZ', 2), ('DT', 3), ('RBS', 4), ('JJ', 5), ('NNS', 6))),
        ('question', ('average'), (('WP', 0), ('VBZ', 1), ('DT', 2), ('JJ', 3), ('NN', 4), ('IN', 5), ('NNP', 6))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NNP', 1), ('NNP', 2))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NN', 1), ('NNP', 2))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NNP', 1), ('NNP', 2))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NN', 1), ('NN', 2))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NNP', 1), ('NNP', 2))),
        ('direction', ('explain', 'describe'), (('NN', 0), ('NN', 1), ('NNP', 2))),
        ('direction', ('explain', 'describe'), (('NN', 0), ('NNP', 1))),
        ('direction', ('explain', 'describe'), (('VB', 0), ('NNP', 1))),
        ('direction', ('explain', 'describe'), (('CD', 0), ('JJ', 1))),
        ('direction', ('explain', 'describe'), (('NNP', 0), ('NNP', 1))),
        ('statement', ('today', 'yesterday', 'tomorrow'), (('MD', 0), ('PRP', 1), ('VB', 2), ('NN', 3))),
        ('question', ('biggest', 'smallest'), (('JJS', 0), ('NN', 1))),
    ]

    DIRECTION = ("describe", "explain", "show", "view")

    NOUN_PLURAL = [
        "There are {} in any organisation. ",
        "Let's keep focus and improve the {}.",
        "yesterdays successes can become {} problems. Keep fighting the good fight."
    ]

    SELF_REFERENCING = [
        "I'm busy keeping the data clean.",
        "I want to help you work smart!",
        "I see widgets in my sleep.",
        "Ready to optimise, optimise, optimise.",
        "I think there is work to be done."
    ]

    NONE_RESPONSES = [
        ["It's probably time for a coffee!"],
        ["It's probably time for some tea!"],
        ["Don't let the details get you down"],
        ["Keep fighting the good fight"],

    ]

    RANDOM_STATEMENTS = [
        "I see SKUs in my sleep",
    ]
    _communication_id = ''
    _currency_symbol = ''
    _currency_code = ''

    def __init__(self):
        self._master_sku_list = get_master_sku_list()
        self.compiled_response = ResponseSingleton()

    @property
    def salutations(self):
        return self.SALUTATION_RESPONSES

    @property
    def currency_symbol(self):
        return self._currency_symbol

    @property
    def currency(self):
        return currency_symbol_controller(database_connection_uri(retrieve='retrieve'))

    @property
    def communication_id(self):
        return self._communication_id

    @communication_id.setter
    def communication_id(self, comms_id):
        self._communication_id = comms_id

    def load_currency_symbol(self):
        self._currency_code = self.currency
        cur = Currency(self._currency_code)
        self._currency_symbol = cur.retrieve_symbol()

    @strip_punctuation
    def _check_for_greeting(self, sentence: str):
        """ Checks whether message includes greeting and responds in kind.

        Args:
            sentence:   Sentence from message entered by user.

        Returns:

        """
        greeting = [random.choice(self.SALUTATION_RESPONSES) for word in sentence.split(' ') if
                    word.lower() in self.SALUTATION_KEYWORDS]
        if len(greeting) == 0:
            greeting = None
            return greeting
        return greeting

    def _check_for_bot_reference(self, pronoun, noun, adjective, verb):
        """

        Args:
            pronoun:
            noun:
            adjective:
            verb:

        Returns:

        """
        if pronoun:
            if pronoun[0] == 'You':
                if noun and adjective is False:
                    response = random.choice(self._NOUN_SINGULAR).format()
                elif noun and verb:
                    response = random.choice(self.NOUN_PLURAL).format(
                        noun[0].pluralize())
                else:
                    response = random.choice(self.SELF_REFERENCING)

                return response
        else:
            return None

    def direction_describe(self, tags):
        """

        Args:
            tags:

        Returns:

        """
        for i in self._master_sku_list:
            if i[1] == tags[1][0]:
                return ['<a href=\"/sku_detail/{}\">Here you go!</a>'.format(i[0])]

    def _understand_direction(self, tags):
        """

        Args:
            tags:

        Returns:

        """
        # understand the direction
        return self.direction_describe(tags)

    def _understand_question(self, dependency, tags, sentence):
        """

        Args:
            dependency:
            tags:
            sentence:

        Returns:

        """
        # understand what the question is asking
        for tag in tags:
            if tag[1].lower() == 'jjs' or tag[1].lower() == 'rbs':
                return self.check_min_max_semantic(dependency, tag[0], sentence)
        for tag in tags:
            if tag[1].lower() == 'nn':
                return self.check_keyword_sku_combo(dependency, tag[0], sentence)

    def check_keyword_sku_combo(self, dependency, nn: str, sentence):

        len_tags = len(sentence.tags)
        new_tags = sentence.tags
        antepenultimate_word = new_tags[len_tags - 3]
        end_word = new_tags[len_tags - 1]
        skus = [sku[1] for sku in  self._master_sku_list]
        if sentence.tags[len_tags - 1][1] in ('NNP') and end_word[0] in skus:
            for word in self.ANALYSIS_KEYWORDS:
                if word == 'classification':
                    result = classification_controller(database_connection_uri(retrieve='retrieve'),sku_id=end_word[0])
                    response = [
                        'SKU {} has the classification {}'.format(str(end_word[0]), result)]
                    return response

    def check_min_max_semantic(self, dependency, jjs: str, sentence):
        """

        Args:
            dependency:
            jjs:
            sentence:

        Returns:

        """
        # probably beter to change this to a tuple of keywords, removes the call to wordnet etc
        max = dependency[1][0]
        min_w = dependency[1][1]
        jj_word = Word(jjs)
        max_word = Word(max)
        min_word = Word(min_w)
        len_tags = len(sentence.tags)
        new_tags = sentence.tags
        penultimate_word = new_tags[len_tags - 2]
        end_word = new_tags[len_tags - 1]

        for word in jj_word.synsets:
            for set in min_word.synsets:
                if word == set:
                    for keyword in self.ANALYSIS_KEYWORDS:
                        if keyword == end_word[0].lower() and keyword == 'excess':
                            result = excess_controller(database_connection_uri(retrieve='retrieve'),
                                                       direction='smallest')
                            response = [
                                'SKU {} has the smallest excess value at {}{:,.2f}'.format(str(result[1]),
                                                                                           self.currency_symbol,
                                                                                           result[0][1])]
                            return response

                        elif keyword == end_word[0].lower() and keyword in ('shortage', 'shortages'):
                            result = shortage_controller(database_connection_uri(retrieve='retrieve'),
                                                         direction='smallest')
                            response = [
                                'SKU {} has the smallest shortage value at {}{:,.2f}'.format(str(result[1]),
                                                                                             self.currency_symbol,
                                                                                             result[0][1])]
                            return response

                        elif keyword == end_word[0].lower() and keyword == 'revenue':
                            result = revenue_controller(database_connection_uri(retrieve='retrieve'),
                                                        direction='smallest')
                            response = [
                                'SKU {} has the lowest revenue value at {}{:,.2f}'.format(str(result[1]),
                                                                                          self.currency_symbol,
                                                                                          result[0][1])]
                            return response
                        elif keyword == '{} {}'.format(penultimate_word[0].lower(),
                                                       end_word[0].lower()) and keyword == 'inventory turns':
                            result = inventory_turns_controller(database_connection_uri(retrieve='retrieve'),
                                                                direction='smallest')
                            response = [
                                'SKU {} has the fewest inventory turns at {}'.format(str(result[1]), result[0][1])]
                            return response
                        elif keyword == '{} {}'.format(penultimate_word[0].lower(),
                                                       end_word[0].lower()) and keyword in ('average order', 'average orders'):
                            result = average_orders_controller(database_connection_uri(retrieve='retrieve'),
                                                               direction='smallest')
                            response = [
                                'SKU {} has the lowest average order at {} units.'.format(str(result[1]),
                                                                                          result[0][1])]
                            return response

                        elif keyword == '{} {}'.format(penultimate_word[0].lower(),
                                                    end_word[0].lower()) and keyword == 'safety stock':
                            result = safety_stock_controller(database_connection_uri(retrieve='retrieve'),
                                                               direction='smallest')
                            response = ['SKU {} has the lowest safety stock at {} units.'.format(str(result[1]),
                                                                                                   result[0])]
                            return response

                        elif keyword == '{} {}'.format(penultimate_word[0].lower(),
                                                    end_word[0].lower()) and keyword == 'reorder level':
                            result = reorder_level_controller(database_connection_uri(retrieve='retrieve'),
                                                               direction='smallest')
                            response = ['SKU {} has the lowest reorder level at {} units.'.format(str(result[1]),
                                                                                                   result[0])]
                            return response

        # print(sentence.tags[len_tags - 1][1])
        if sentence.tags[len_tags - 1][1] in ('NN') or max.lower() in self._MAX:
            for word in self.ANALYSIS_KEYWORDS:
                if word == end_word[0] and word == 'excess':
                    # max excess filters excess rank for lowest rank indicating biggest excess
                    result = excess_controller(database_connection_uri(retrieve='retrieve'), direction='biggest')
                    return [
                        'SKU {} has the largest excess value at {}{:,.2f}'.format(str(result[1]), self.currency_symbol,
                                                                                  result[0][1])]
                elif word == end_word[0].lower() and word == 'shortage':
                    result = shortage_controller(database_connection_uri(retrieve='retrieve'), direction='biggest')
                    return ['SKU {} has the largest shortage value at {}{:,.2f}'.format(str(result[1]),
                                                                                        self.currency_symbol,
                                                                                        result[0][1])]
                elif word == end_word[0].lower() and word == 'revenue':
                    result = revenue_controller(database_connection_uri(retrieve='retrieve'), direction='biggest')
                    return [
                        'SKU {} has the highest revenue value at {}{:,.2f}'.format(str(result[1]), self.currency_symbol,
                                                                                   result[0][1])]
                elif word == '{} {}'.format(penultimate_word[0].lower(),
                                            end_word[0].lower()) and word == 'inventory turns':
                    result = inventory_turns_controller(database_connection_uri(retrieve='retrieve'),
                                                        direction='biggest')
                    response = ['SKU {} has the highest inventory turn rate at {}'.format(str(result[1]), result[0][1])]
                    return response
                elif word == '{} {}'.format(penultimate_word[0].lower(),
                                            end_word[0].lower()) and word in ('average order', 'average orders'):
                    result = average_orders_controller(database_connection_uri(retrieve='retrieve'),
                                                       direction='biggest')
                    response = ['SKU {} has the largest average order at {} units.'.format(str(result[1]), result[0][1])]
                    return response
                elif word == '{} {}'.format(penultimate_word[0].lower(),
                                                end_word[0].lower()) and word =='safety stock':
                    result = safety_stock_controller(database_connection_uri(retrieve='retrieve'),
                                                       direction='biggest')
                    response = [
                        'SKU {} has the largest safety stock at {} units.'.format(str(result[1]), result[0])]
                    return response
                elif word == '{} {}'.format(penultimate_word[0].lower(),
                                                end_word[0].lower()) and word == 'reorder level':
                    result = reorder_level_controller(database_connection_uri(retrieve='retrieve'),
                                                     direction='biggest')
                    response = [
                        'SKU {} has the largest reorder level at {} units.'.format(str(result[1]), result[0])]
                    return response


                    # print(max_word.synsets, min_word.synsets, 'jj: {} '.format(jj_word.synsets))

    def _deconstruct_message(self, sentence) -> list:
        """

        Args:
            sentence:

        Returns:

        """
        # deconstruct using tree and classifier then select path
        new_tags = tuple([(i[1], index) for index, i in enumerate(sentence.tags)])
        for i in self._SEMANTIC_DICTIONARY:
            if i[2] == new_tags:
                if i[0] == 'question':
                    return self._understand_question(i, sentence.tags, sentence)
                elif i[0] == 'direction':
                    return self._understand_direction(sentence.tags)
                elif i[0] == 'instruction':
                    pass
                elif i[0] == 'statement':
                    pass
                    # print(i, sentence.tags, sentence)
                else:
                    return None

    def _append_response(self, message):
        msg_flag = ''
        msg = []
        if self.communication_id:
            msg_flag = self.compiled_response.shared_response.get(str(self.communication_id), self._EMPTY)
        else:
            msg_flag = self._EMPTY
            self.communication_id = uuid.uuid1()
        if self._EMPTY != msg_flag:
            msg.append(msg_flag)
            for m in message:
                msg.append(m)

            self.compiled_response.shared_response.update(**{'{}'.format(self.communication_id): msg})
        else:
            self.compiled_response.shared_response.update(**{'{}'.format(self.communication_id): message})

    def _check_deconstruction_required(self, message: dict) -> bool:
        sentence_count = len([i for i in message.keys()])
        # precompiled_response = [response for response in self.compiled_response.shared_response.values()]
        if sentence_count != 0:
            return True
        else:
            return False

    def unpack_pos(self, msg):
        pronoun = None
        noun = None
        adjective = None
        verb = None
        pronoun = _find_pronoun(filter_pos(sentence=msg, pos='PRP'))
        noun = filter_pos_type(sentence=msg, pos_type='noun')
        adjective = filter_pos_type(sentence=msg, pos_type='adjective')
        verb = filter_pos_type(sentence=msg, pos_type='verb')
        # print("pronoun: {}, noun: {}, adjective: {}, verb {},".format(pronoun, noun, adjective, verb))
        return {'pronoun': pronoun, 'noun': noun, 'adjective': adjective, 'verb': verb}

    def initialise_chat(self, message):
        u_wot_m8 = TextBlob(message)
        self.load_currency_symbol()
        unpacked_sentences = {}
        new_state = ''
        for sentence in u_wot_m8.sentences:
            unpacked_sentence_pos = _unpack_pos(message=sentence)
            unpacked_sentences.update({sentence: unpacked_sentence_pos})
            new_state = self._TRANSITION_STATES.get('PRONOUN_STATE', self._END_STATE)
        return new_state, unpacked_sentences

    def pronoun(self, message: dict):
        msg = {}
        for x in message.values():
            msg.update(**x)
        response = self._check_for_bot_reference(**msg)
        if response:
            new_state = self._TRANSITION_STATES.get('DECONSTRUCTION_STATE', self._END_STATE)
            self._append_response(message=response)
        else:
            new_state = self._TRANSITION_STATES.get('GREETING_STATE', self._END_STATE)
        return new_state, message

    def greeting(self, message: dict):
        msg = []
        for x in message.keys():
            msg.append(x)
        response = ''
        del_key = ''
        for sentence in msg:
            response = self._check_for_greeting(*msg)
            if response:
                del_key = sentence
                break
        new_state = self._TRANSITION_STATES.get('DECONSTRUCTION_STATE', self._END_STATE)
        if response:
            self._append_response(message=response)
            del message[del_key]
        return new_state, message

    def deconstruction(self, message: dict):
        msg = []
        response = []
        new_state = new_state = self._TRANSITION_STATES.get('RANDOM_STATE', self._END_STATE)
        for x in message.keys():
            msg.append(x)
        if self._check_deconstruction_required(message=message):
            reply = []
            for sentence in message.keys():
                response = self._deconstruct_message(sentence)
                if response:
                    reply.append(response)
                    new_state = self._TRANSITION_STATES.get('RESPONSE_STATE', self._END_STATE)
            if reply:
                self._append_response(message=reply)
            else:
                new_state = new_state = self._TRANSITION_STATES.get('RANDOM_STATE', self._END_STATE)
        return new_state, message

    def random_utterance(self, message: dict):
        response = ''
        for sentence in message.keys():
            print(sentence)
            unpacked_sentence_pos = self.unpack_pos(sentence)
            if not unpacked_sentence_pos['pronoun']:
                response = random.choice(self.NONE_RESPONSES)
            elif unpacked_sentence_pos['pronoun'] == 'I' and not unpacked_sentence_pos['verb'][0]:
                response = random.choice(self.RANDOM_STATEMENTS)
        self._append_response(message=response)
        new_state = self._TRANSITION_STATES.get('RESPONSE_STATE', self._END_STATE)
        return new_state, message



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal
from supplychainpy import model_inventory
from supplychainpy._helpers._pickle_config import deserialise_config
from supplychainpy.bi._recommendation_state_machine import SkuMachine, ProfileMachine
from supplychainpy.bi._recommendations import SKUStates, ProfileStates
from supplychainpy.inventory.analyse_uncertain_demand import UncertainDemand
from supplychainpy.sample_data.config import ABS_FILE_PATH, FORECAST_PICKLE


def run_sku_recommendation(analysed_orders:UncertainDemand, forecast: dict)->dict:
    """ Runs SKU recommendation state machine and generates recommendations for each sku.

    Args:
        analysed_orders (UncertainDemand):  Analysed Orders.
        forecast (dict):                    forecast.

    Returns:
        dict:   Recommendations for each sku.

    """
    recommend = SkuMachine()
    states = SKUStates(analysed_orders=analysed_orders, forecast=forecast)
    recommend.add_state("start", states.initialise_machine)
    recommend.add_state("excess_rank", states.excess_rank)
    recommend.add_state("shortage_rank", states.shortage_rank)
    recommend.add_state("inventory_turns", states.inventory_turns)
    recommend.add_state("classification", states.classification)
    recommend.add_state("traffic_light", states.traffic_light)
    recommend.add_state("forecast", states.forecast)
    recommend.add_state("recommendation", recommend, end_state=1)
    recommend.set_start("start")
    for sku in analysed_orders:
        recommend.run(sku.sku_id)
    return deserialise_config(ABS_FILE_PATH['RECOMMENDATION_PICKLE'])


def run_profile_recommendation(analysed_orders: UncertainDemand, forecast: dict)->dict:
    """ Runs profile recommendation state machine and generates recommendations for entire inventory profile.

    Args:
        analysed_orders (UncertainDemand):  Analysed orders as UncertainDemand object
        forecast (dict):                    Forecast results for same data set.

    Returns:
        dict:    Recommendations for entire inventory profile

    """
    recommend = ProfileMachine()
    states = ProfileStates(analysed_orders=analysed_orders, forecast=forecast)
    recommend.add_state("start", states.initialise_machine)
    recommend.add_state("revenue", states.revenue)
    recommend.add_state("excess", states.excess)
    recommend.add_state("shortage", states.shortage)
    recommend.add_state("classification", states.classification)
    recommend.add_state("inventory", states.inventory_turns)
    recommend.add_state("recommendation", recommend, end_state=1)
    recommend.set_start("start")
    recommend.run()
    # have to serialise the profile recommendation seperately from the sku recommendation
    return deserialise_config(ABS_FILE_PATH['PROFILE_PICKLE'])


if __name__ == '__main__':
    orders_analysis = model_inventory.analyse(file_path=ABS_FILE_PATH['COMPLETE_CSV_LG'],
                                              z_value=Decimal(1.28),
                                              reorder_cost=Decimal(5000),
                                              file_type="csv",
                                              length=12)

    # d = ProfileGenerator(analysed_orders=orders_analysis)
    # d.Top_Concerns()
    # resp = {}
    for i in run_sku_recommendation(analysed_orders=orders_analysis, forecast=deserialise_config(FORECAST_PICKLE)).values():
       print(i)
    # d = ProfileStates(analysed_orders=orders_analysis, forecast=deserialise_config(FORECAST_PICKLE))
    # d.revenue()
    #for i in run_profile_recommendation(analysed_orders=orders_analysis,
    #                                    forecast=deserialise_config(FORECAST_PICKLE)).values():
    #    print(i)



import random
import re

from supplychainpy._helpers._db_connection import database_connection_uri
from supplychainpy._helpers._decorators import strip_punctuation
from supplychainpy.bot._controller import master_sku_list
from textblob import TextBlob
from textblob import Word


def get_master_sku_list():
    msk = master_sku_list(database_connection_uri(retrieve='retrieve'))
    return msk





def filter_pos(sentence, pos):
    """

    Args:
        sentence:
        pos:

    Returns:

    """
    tags = [i for i in sentence.tags]
    for tag in tags:
        if tag[1] == pos:
            return tag


def filter_pos_type(sentence, pos_type):
    """

    Args:
        sentence:
        pos_type:

    Returns:

    """
    pos_regex = ''
    if pos_type == 'verb':
        pos_regex = re.compile('[V][B]\w?\w?')
    if pos_type == 'adjective':
        pos_regex = re.compile('[J][J]\w?\w?')
    if pos_type == 'noun':
        pos_regex = re.compile('[N][N]\w?\w?')

    tags = [i for i in sentence.tags]
    for tag in tags:
        if pos_regex.match(tag[1]):
            return tag


def _unpack_pos( message):
    pronoun = _find_pronoun(filter_pos(sentence=message, pos='PRP'))
    noun = filter_pos_type(sentence=message, pos_type='noun')
    adjective = filter_pos_type(sentence=message, pos_type='adjective')
    verb = filter_pos_type(sentence=message, pos_type='verb')

    # print("pronoun: {}, noun: {}, adjective: {}, verb {},".format(pronoun, noun, adjective, verb))
    return {'pronoun': pronoun, 'noun': noun, 'adjective': adjective, 'verb': verb}


def _find_pronoun(pos: tuple) -> str:
    """

    Args:
        pos:

    Returns:

    """
    if pos is not None:
        pronoun = 'I' if pos[0] == 'PRP' and pos[1].lower() == 'you' else 'You'
        return pronoun
    else:
        return None


def unpack_sentence(message):
    """

    Args:
        message:

    Returns:

    """
    u_wot_m8 = TextBlob(message)
    print(u_wot_m8.tags)
    for i in u_wot_m8.tags:
        syn = Word(i[0])
        print('word: {}\n lemmatize: {}\n lemma: {}\n synsets: {}\n'.format(i[0], syn.lemmatize, syn.lemma,
                                                                            syn.synsets))


